"""This code includes some models of jos-3 model to calculate human
thermoregulation.

The values of a NumPy array containing 17 body parts
"""

import numpy as np

from pythermalcomfort.classes_return import JOS3BodyParts
from pythermalcomfort.jos3_functions import construction as cons
from pythermalcomfort.jos3_functions.matrix import IDICT
from pythermalcomfort.jos3_functions.parameters import Default
from pythermalcomfort.utilities import Postures, Sex, antoine


def natural_convection(posture: str, tdb: float, t_skin: float) -> np.ndarray:
    """Calculate the natural convection heat transfer coefficient based on posture.

    Parameters
    ----------
    posture : str
        The posture of the subject. Valid options are "standing", "sitting", "lying", "sedentary", "supine".
    tdb : float
        Dry bulb temperature [°C].
    t_skin : float
        Skin temperature [°C].

    Returns
    -------
    hc_natural : numpy.ndarray
        Natural convection heat transfer coefficient body segments.

    Raises
    ------
    ValueError
        If the posture is not one of the valid options.

    """
    if posture.lower() == Postures.standing.value:
        # Ichihara et al., 1997, https://doi.org/10.3130/aija.62.45_5
        hc_natural = np.array(
            [
                4.48,
                4.48,
                2.97,
                2.91,
                2.85,
                3.61,
                3.55,
                3.67,
                3.61,
                3.55,
                3.67,
                2.80,
                2.04,
                2.04,
                2.80,
                2.04,
                2.04,
            ],
        )
    elif posture.lower() in [Postures.sitting.value, Postures.sedentary.value]:
        # Ichihara et al., 1997, https://doi.org/10.3130/aija.62.45_5
        hc_natural = np.array(
            [
                4.75,
                4.75,
                3.12,
                2.48,
                1.84,
                3.76,
                3.62,
                2.06,
                3.76,
                3.62,
                2.06,
                2.98,
                2.98,
                2.62,
                2.98,
                2.98,
                2.62,
            ],
        )
    elif posture.lower() in [Postures.lying.value, Postures.supine.value]:
        # Kurazumi et al., 2008, https://doi.org/10.20718/jjpa.13.1_17
        # The values are applied under cold environment.
        hc_a = np.array(
            [
                1.105,
                1.105,
                1.211,
                1.211,
                1.211,
                0.913,
                2.081,
                2.178,
                0.913,
                2.081,
                2.178,
                0.945,
                0.385,
                0.200,
                0.945,
                0.385,
                0.200,
            ],
        )
        hc_b = np.array(
            [
                0.345,
                0.345,
                0.046,
                0.046,
                0.046,
                0.373,
                0.850,
                0.297,
                0.373,
                0.850,
                0.297,
                0.447,
                0.580,
                0.966,
                0.447,
                0.580,
                0.966,
            ],
        )
        hc_natural = hc_a * (abs(tdb - t_skin) ** hc_b)
    else:
        valid_postures = [
            Postures.standing.value,
            Postures.sitting.value,
            Postures.lying.value,
            Postures.sedentary.value,
            Postures.supine.value,
        ]
        error_msg = f"Invalid posture '{posture}'. Must be one of {valid_postures}."
        raise ValueError(error_msg)
    return hc_natural


def forced_convection(v: float) -> np.ndarray:
    """Calculate the forced convection heat transfer coefficient.

    Ichihara et al., 1997, https://doi.org/10.3130/aija.62.45_5

    Parameters
    ----------
    v : float
        Air velocity [m/s].

    Returns
    -------
    np.ndarray
        Forced convection heat transfer coefficient body segments.

    """
    hc_a = np.array(
        [
            15.0,
            15.0,
            11.0,
            17.0,
            13.0,
            17.0,
            17.0,
            20.0,
            17.0,
            17.0,
            20.0,
            14.0,
            15.8,
            15.1,
            14.0,
            15.8,
            15.1,
        ],
    )
    hc_b = np.array(
        [
            0.62,
            0.62,
            0.67,
            0.49,
            0.60,
            0.59,
            0.61,
            0.60,
            0.59,
            0.61,
            0.60,
            0.61,
            0.74,
            0.62,
            0.61,
            0.74,
            0.62,
        ],
    )
    hc_forced = hc_a * (v**hc_b)
    return hc_forced


def conv_coef(
    posture: str,
    v: float | np.ndarray,
    tdb: float | np.ndarray,
    t_skin: float | np.ndarray,
) -> np.ndarray:
    """Calculate convective heat transfer coefficient (hc) [W/(m2*K)].

    Ichihara et al., 1997, https://doi.org/10.3130/aija.62.45_5
    Kurazumi et al., 2008, https://doi.org/10.20718/jjpa.13.1_17

    Parameters
    ----------
    posture : str, optional
        Select posture from standing, sitting, lying, sedentary or supine.
    v : float or iter, optional
        Air velocity [m/s]. If iter is input, its length should be 17.
        The default is 0.1.
    tdb : float or iter, optional
        Air temperature [°C]. If iter is input, its length should be 17.
        The default is 28.8.
    t_skin : float or iter, optional
        Skin temperature [°C]. If iter is input, its length should be 17.
        The default is 34.0.

    Returns
    -------
    np.ndarray
        Convective heat transfer coefficient (hc) [W/(m2*K)].

    """
    hc_natural = natural_convection(posture=posture, tdb=tdb, t_skin=t_skin)
    hc_forced = forced_convection(v=v)
    hc = np.where(v < 0.2, hc_natural, hc_forced)  # hc [W/(m2*K))]
    return hc


def rad_coef(posture: str) -> np.ndarray:
    """Calculate radiative heat transfer coefficient (hr) [W/(m2*K)].

    Parameters
    ----------
    posture : str, optional
        Select posture from standing, sitting, lying, sedentary or supine.

    Returns
    -------
    np.ndarray
        Radiative heat transfer coefficient (hr) [W/(m2*K)].

    """
    if posture.lower() == Postures.standing.value:
        # Ichihara et al., 1997, https://doi.org/10.3130/aija.62.45_5
        hr = np.array(
            [
                4.89,
                4.89,
                4.32,
                4.09,
                4.32,
                4.55,
                4.43,
                4.21,
                4.55,
                4.43,
                4.21,
                4.77,
                5.34,
                6.14,
                4.77,
                5.34,
                6.14,
            ],
        )
    elif posture.lower() in [Postures.sitting.value, Postures.sedentary.value]:
        # Ichihara et al., 1997, https://doi.org/10.3130/aija.62.45_5
        hr = np.array(
            [
                4.96,
                4.96,
                3.99,
                4.64,
                4.21,
                4.96,
                4.21,
                4.74,
                4.96,
                4.21,
                4.74,
                4.10,
                4.74,
                6.36,
                4.10,
                4.74,
                6.36,
            ],
        )
    elif posture.lower() in [Postures.lying.value, Postures.supine.value]:
        # Kurazumi et al., 2008, https://doi.org/10.20718/jjpa.13.1_17
        hr = np.array(
            [
                5.475,
                5.475,
                3.463,
                3.463,
                3.463,
                4.249,
                4.835,
                4.119,
                4.249,
                4.835,
                4.119,
                4.440,
                5.547,
                6.085,
                4.440,
                5.547,
                6.085,
            ],
        )
    else:
        valid_postures = [
            Postures.standing.value,
            Postures.sitting.value,
            Postures.lying.value,
            Postures.sedentary.value,
            Postures.supine.value,
        ]
        error_msg = f"Invalid posture '{posture}'. Must be one of {valid_postures}."
        raise ValueError(
            error_msg,
        )
    return hr


def fixed_hc(hc: np.ndarray, v: np.ndarray) -> np.ndarray:
    """Fix hc values to fit two-node-model's values.

    Parameters
    ----------
    hc : np.ndarray
        Convective heat transfer coefficient (hc) [W/(m2*K)].
    v : float
        Air velocity [m/s].

    Returns
    -------
    np.ndarray
        Fixed convective heat transfer coefficient (hc) [W/(m2*K)].

    """
    mean_hc = np.average(hc, weights=cons.Default.local_bsa)
    mean_va = np.average(v, weights=cons.Default.local_bsa)
    mean_hc_whole = max(3, 8.600001 * (mean_va**0.53))
    _fixed_hc = hc * mean_hc_whole / mean_hc
    return _fixed_hc


def fixed_hr(hr: np.ndarray) -> np.ndarray:
    """Fix hr values to fit two-node-model's values.

    Parameters
    ----------
    hr : np.ndarray
        Radiative heat transfer coefficient (hr) [W/(m2*K)].

    Returns
    -------
    np.ndarray
        Fixed radiative heat transfer coefficient (hr) [W/(m2*K)].

    """
    mean_hr = np.average(hr, weights=cons.Default.local_bsa)
    _fixed_hr = hr * 4.7 / mean_hr
    return _fixed_hr


# TODO this function is a duplicate in utils
def operative_temp(tdb, tr, hc, hr):
    """Calculate operative temperature [°C].

    Parameters
    ----------
    tdb : float or array
        Air temperature [°C]
    tr : float or array
        Mean radiant temperature [°C]
    hc : float or array
        Convective heat transfer coefficient [W/(m2*K)]
    hr : float or array
        Radiative heat transfer coefficient [W/(m2*K)]

    Returns
    -------
    to : float or array
        Operative temperature [°C]

    """
    to = (hc * tdb + hr * tr) / (hc + hr)
    return to


# TODO this function is a duplicate in utils and they are different
def clo_area_factor(clo):
    """Calculate clothing area factor.

    Parameters
    ----------
    clo : float or array
        Clothing insulation [clo]

    Returns
    -------
    fcl : float or array
        clothing area factor [-]

    """
    fcl = np.where(clo < 0.5, clo * 0.2 + 1, clo * 0.1 + 1.05)
    return fcl


# TODO this function is different from ISO 9920
def dry_r(hc, hr, clo):
    """Calculate total sensible thermal resistance (between the skin and
    ambient air).

    Parameters
    ----------
    hc : float or array
        Convective heat transfer coefficient (hc) [W/(m2*K)].
    hr : float or array
        Radiative heat transfer coefficient (hr) [W/(m2*K)].
    clo : float or array
        Clothing insulation [clo].

    Returns
    -------
    np.ndarray
        Total sensible thermal resistance between skin and ambient.

    Raises
    ------
    ValueError
        If any of the input parameters are negative.

    """
    if (np.array(hc) < 0).any() or (np.array(hr) < 0).any():
        raise ValueError("Input parameters hc and hr must be non-negative.")

    fcl = clo_area_factor(clo)
    r_a = 1 / (hc + hr)
    r_cl = 0.155 * clo
    r_t = r_a / fcl + r_cl
    return r_t


def wet_r(
    hc,
    clo,
    i_clo,
    lewis_rate,
):
    """Calculate total evaporative thermal resistance (between the skin and
    ambient air).

    Parameters
    ----------
    hc : float or array
        Convective heat transfer coefficient (hc) [W/(m2*K)].
    clo : float or array
        Clothing insulation [clo].
    i_clo : float or array, optional
        Clothing vapor permeation efficiency [-]. The default is 0.45.
    lewis_rate : float, optional
        Lewis rate [K/kPa]. The default is 16.5.

    Returns
    -------
    np.ndarray
        Total evaporative thermal resistance.

    Raises
    ------
    ValueError
        If any of the input parameters are negative.

    """
    if (np.array(hc) < 0).any():
        raise ValueError("Input parameters hc must be non-negative.")

    fcl = clo_area_factor(clo)
    r_cl = 0.155 * clo
    r_ea = 1 / (lewis_rate * hc)
    r_ecl = r_cl / (lewis_rate * i_clo)
    r_et = r_ea / fcl + r_ecl
    return r_et


def error_signals(err_sk=0.0):
    """Calculate WRMS and CLDS signals of thermoregulation.

    Parameters
    ----------
    err_sk : float or np.ndarray, optional
        Difference between set-point and skin temperatures [°C]. The default is 0.

    Returns
    -------
    tuple
        Warm signal (WRMS) [°C] and Cold signal (CLDS) [°C].

    """
    err_sk = np.array(err_sk, dtype=float)

    # SKINR (Distribution coefficients of thermal receptor) [-]
    receptor = np.array(
        [
            0.0549,
            0.0146,
            0.1492,
            0.1321,
            0.2122,
            0.0227,
            0.0117,
            0.0923,
            0.0227,
            0.0117,
            0.0923,
            0.0501,
            0.0251,
            0.0167,
            0.0501,
            0.0251,
            0.0167,
        ],
    )

    # wrms signal
    wrm = np.maximum(err_sk, 0) * receptor
    warm_signal_sum = wrm.sum()

    # clds signal
    cld = np.minimum(err_sk, 0) * -receptor
    cold_signal_sum = cld.sum()
    return warm_signal_sum, cold_signal_sum


def tetens(x: float) -> float:
    """Calculate saturated vapor pressure using Tetens equation [kPa].

    Parameters
    ----------
    x : float
        Temperature [°C].

    Returns
    -------
    float
        Saturated vapor pressure [kPa].

    """
    return 0.61078 * 10 ** (7.5 * x / (x + 237.3))


def evaporation(
    err_cr,
    err_sk,
    t_skin,
    tdb,
    rh,
    ret,
    height,
    weight,
    bsa_equation,
    age,
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Calculate evaporative heat loss.

    Parameters
    ----------
    err_cr, err_sk : array
        Difference between set-point and body temperatures [°C].
    t_skin : array
        Skin temperatures [°C].
    tdb : array
        Air temperatures at local body segments [°C].
    rh : array
        Relative humidity at local body segments [%].
    ret : array
        Total evaporative thermal resistances [m2.K/W].
    height : float, optional
        Body height [m].
    weight : float, optional
        Body weight [kg].
    bsa_equation : str, optional
        The equation name (str) of bsa calculation. Choose one from BodySurfaceAreaEquations.
    age : float, optional
        age [years].

    Returns
    -------
    wet : array
        Local skin wettedness [-].
    e_sk : array
        Evaporative heat loss at the skin by sweating and diffuse [W].
    e_max : array
        Maximum evaporative heat loss at the skin [W].
    e_sweat : TYPE
        Evaporative heat loss at the skin by only sweating [W].

    """
    wrms, clds = error_signals(
        err_sk,
    )  # Thermoregulation signals
    bsar = cons.bsa_rate(
        height,
        weight,
        bsa_equation,
    )  # bsa rate
    bsa = Default.local_bsa * bsar  # bsa
    p_a = antoine(tdb) * rh / 100  # Saturated vapor pressure of ambient [kPa]
    p_sk_s = antoine(t_skin)  # Saturated vapor pressure at the skin [kPa]

    e_max = (p_sk_s - p_a) / ret * bsa  # Maximum evaporative heat loss

    # Replace any zero values in the e_max array with 0.001 to avoid causing a divide by 0 error
    e_max = np.where(e_max == 0, 0.001, e_max)

    # SKINS
    skin_sweat = np.array(
        [
            0.064,
            0.017,
            0.146,
            0.129,
            0.206,
            0.051,
            0.026,
            0.0155,
            0.051,
            0.026,
            0.0155,
            0.073,
            0.036,
            0.0175,
            0.073,
            0.036,
            0.0175,
        ],
    )

    sig_sweat = (371.2 * err_cr[0]) + (33.64 * (wrms - clds))
    sig_sweat = max(sig_sweat, 0)
    sig_sweat *= bsar

    # Signal decrement by aging
    if age < 60:
        sd_sweat = np.ones(Default.num_body_parts)
    else:  # age >= 60
        sd_sweat = np.array(
            [
                0.69,
                0.69,
                0.59,
                0.52,
                0.40,
                0.75,
                0.75,
                0.75,
                0.75,
                0.75,
                0.75,
                0.40,
                0.40,
                0.40,
                0.40,
                0.40,
                0.40,
            ],
        )
    e_sweat = skin_sweat * sig_sweat * sd_sweat * 2 ** (err_sk / 10)
    wet = 0.06 + 0.94 * (e_sweat / e_max)
    wet = np.minimum(wet, 1)  # Wettedness' upper limit
    e_sk = wet * e_max
    e_sweat = (wet - 0.06) / 0.94 * e_max  # Effective sweating
    return wet, e_sk, e_max, e_sweat


def skin_blood_flow(
    err_cr,
    err_sk,
    height,
    weight,
    bsa_equation,
    age,
    ci,
) -> np.ndarray:
    """Calculate skin blood flow rate (bf_skin) [L/h].

    Parameters
    ----------
    err_cr, err_sk : array
        Difference between set-point and body temperatures [°C].
    height : float, optional
        Body height [m].
    weight : float, optional
        Body weight [kg].
    bsa_equation : str, optional
        The equation name (str) of bsa calculation. Choose one from BodySurfaceAreaEquations.
    age : float, optional
        age [years].
    ci : float, optional
        Cardiac index [L/min/㎡].

    Returns
    -------
    bf_skin : array
        Skin blood flow rate [L/h].

    """
    wrms, clds = error_signals(err_sk)

    # BFBsk
    bfb_sk = np.array(
        [
            1.754,
            0.325,
            1.967,
            1.475,
            2.272,
            0.91,
            0.508,
            1.114,
            0.91,
            0.508,
            1.114,
            1.456,
            0.651,
            0.934,
            1.456,
            0.651,
            0.934,
        ],
    )
    # SKIND
    skin_dilat = np.array(
        [
            0.0692,
            0.0992,
            0.0580,
            0.0679,
            0.0707,
            0.0400,
            0.0373,
            0.0632,
            0.0400,
            0.0373,
            0.0632,
            0.0736,
            0.0411,
            0.0623,
            0.0736,
            0.0411,
            0.0623,
        ],
    )
    # SKINC
    skin_stric = np.array(
        [
            0.0213,
            0.0213,
            0.0638,
            0.0638,
            0.0638,
            0.0213,
            0.0213,
            0.1489,
            0.0213,
            0.0213,
            0.1489,
            0.0213,
            0.0213,
            0.1489,
            0.0213,
            0.0213,
            0.1489,
        ],
    )

    sig_dilat = (100.5 * err_cr[0]) + (6.4 * (wrms - clds))
    sig_stric = (-10.8 * err_cr[0]) + (-10.8 * (wrms - clds))
    sig_dilat = max(sig_dilat, 0)
    sig_stric = max(sig_stric, 0)

    # Signal decrement by aging
    if age < 60:
        sd_dilat = np.ones(Default.num_body_parts)
        sd_stric = np.ones(Default.num_body_parts)
    else:  # age >= 60
        sd_dilat = np.array(
            [
                0.91,
                0.91,
                0.47,
                0.47,
                0.31,
                0.47,
                0.47,
                0.47,
                0.47,
                0.47,
                0.47,
                0.31,
                0.31,
                0.31,
                0.31,
                0.31,
                0.31,
            ],
        )
        sd_stric = np.ones(Default.num_body_parts)

    # Skin blood flow [L/h]
    bf_skin = (
        (1 + skin_dilat * sd_dilat * sig_dilat)
        / (1 + skin_stric * sd_stric * sig_stric)
        * bfb_sk
        * 2 ** (err_sk / 6)
    )
    # Basal blood flow rate to the standard body [-]
    bfb_rate = cons.bfb_rate(
        height,
        weight,
        bsa_equation,
        age,
        ci,
    )
    bf_skin *= bfb_rate
    return bf_skin


def ava_blood_flow(
    err_cr,
    err_sk,
    height,
    weight,
    bsa_equation,
    age,
    ci,
):
    """Calculate areteriovenous anastmoses (AVA) blood flow rate [L/h] based on
    Takemori's model, 1995.

    Parameters
    ----------
    err_cr, err_sk : array
        Difference between set-point and body temperatures [°C].
    height : float, optional
        Body height [m].
    weight : float, optional
        Body weight [kg].
    bsa_equation : str, optional
        The equation name (str) of bsa calculation. Choose one from BodySurfaceAreaEquations.
    age : float, optional
        age [years].
    ci : float, optional
        Cardiac index [L/min/m2].

    Returns
    -------
    bf_ava_hand, bf_ava_foot : array
        AVA blood flow rate at hand and foot [L/h].

    """
    # Cal. mean error body core temp.
    cap_bcr = [10.2975, 9.3935, 4.488]  # Thermal capacity at chest, back and pelvis
    err_bcr = np.average(err_cr[2:5], weights=cap_bcr)

    # Cal. mean error skin temp.
    bsa = cons.Default.local_bsa
    err_msk = np.average(err_sk, weights=bsa)

    # Openness of AVA [-]
    sig_ava_hand = 0.265 * (err_msk + 0.43) + 0.953 * (err_bcr + 0.1905) + 0.9126
    sig_ava_foot = 0.265 * (err_msk - 0.997) + 0.953 * (err_bcr + 0.0095) + 0.9126

    sig_ava_hand = min(sig_ava_hand, 1)
    sig_ava_hand = max(sig_ava_hand, 0)
    sig_ava_foot = min(sig_ava_foot, 1)
    sig_ava_foot = max(sig_ava_foot, 0)

    # Basal blood flow rate to the standard body [-]
    bfb_rate = cons.bfb_rate(
        height,
        weight,
        bsa_equation,
        age,
        ci,
    )
    # AVA blood flow rate [L/h]
    bf_ava_hand = 1.71 * bfb_rate * sig_ava_hand  # Hand
    bf_ava_foot = 2.16 * bfb_rate * sig_ava_foot  # Foot
    return bf_ava_hand, bf_ava_foot


def basal_met(
    height,
    weight,
    age,
    sex,
    bmr_equation,
):
    """Calculate basal metabolic rate [W].

    Parameters
    ----------
    height : float, optional
        Body height [m]. The default is 1.72.
    weight : float, optional
        Body weight [kg]. The default is 74.43.
    age : float, optional
        age [years]. The default is 20.
    sex : str, optional
        Choose male or female. The default is "male".
    bmr_equation : str, optional
        Choose harris-benedict or ganpule. The default is "harris-benedict".

    Returns
    -------
     bmr: float
        Basal metabolic rate [W].

    """
    if bmr_equation == "harris-benedict":
        if sex == Sex.male.value:
            bmr = 88.362 + 13.397 * weight + 500.3 * height - 5.677 * age
        else:
            bmr = 447.593 + 9.247 * weight + 479.9 * height - 4.330 * age

    elif bmr_equation == "harris-benedict_origin":
        if sex == Sex.male.value:
            bmr = 66.4730 + 13.7516 * weight + 500.33 * height - 6.7550 * age
        else:
            bmr = 655.0955 + 9.5634 * weight + 184.96 * height - 4.6756 * age

    elif bmr_equation == "japanese" or bmr_equation == "ganpule":
        # Ganpule et al., 2007, https://doi.org/10.1038/sj.ejcn.1602645
        if sex == Sex.male.value:
            bmr = 0.0481 * weight + 2.34 * height - 0.0138 * age - 0.4235
        else:
            bmr = 0.0481 * weight + 2.34 * height - 0.0138 * age - 0.9708
        bmr *= 1000 / 4.186
    else:
        # TODO why we cannot use dubois?
        valid_equations = [
            "harris-benedict",
            "harris-benedict_origin",
            "japanese",
            "ganpule",
        ]
        error_msg = (
            f"Invalid equation: '{bmr_equation}'. Must be one of {valid_equations}."
        )
        raise ValueError(error_msg)

    bmr *= 0.048  # [kcal/day] to [W]

    # Set minimum BMR value in W
    min_bmr_in_w = 68
    bmr = max(bmr, min_bmr_in_w)

    return bmr


def local_mbase(
    height: float = Default.height,
    weight: float = Default.weight,
    age: int = Default.age,
    sex: str = Default.sex,
    bmr_equation: str = Default.bmr_equation,
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Calculate local basal metabolic rate [W].

    Parameters
    ----------
    height : float, optional
        Body height [m]. The default is 1.72.
    weight : float, optional
        Body weight [kg]. The default is 74.43.
    age : int, optional
        age [years]. The default is 20.
    sex : str, optional
        Choose male or female.
    bmr_equation : str, optional
        Choose harris-benedict or ganpule. The default is "harris-benedict".

    Returns
    -------
    mbase : array
        Local basal metabolic rate (Mbase) [W].

    """
    mbase_all = basal_met(height, weight, age, sex, bmr_equation)
    # Distribution coefficient of basal metabolic rate
    mbf_cr = np.array(
        [
            0.19551,
            0.00324,
            0.28689,
            0.25677,
            0.09509,
            0.01435,
            0.00409,
            0.00106,
            0.01435,
            0.00409,
            0.00106,
            0.01557,
            0.00422,
            0.00250,
            0.01557,
            0.00422,
            0.00250,
        ],
    )
    mbf_ms = np.array(
        [
            0.00252,
            0.0,
            0.0,
            0.0,
            0.04804,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
    )
    mbf_fat = np.array(
        [
            0.00127,
            0.0,
            0.0,
            0.0,
            0.00950,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
    )
    mbf_sk = np.array(
        [
            0.00152,
            0.00033,
            0.00211,
            0.00187,
            0.00300,
            0.00059,
            0.00031,
            0.00059,
            0.00059,
            0.00031,
            0.00059,
            0.00144,
            0.00027,
            0.00118,
            0.00144,
            0.00027,
            0.00118,
        ],
    )

    mbase_cr = mbf_cr * mbase_all
    mbase_ms = mbf_ms * mbase_all
    mbase_fat = mbf_fat * mbase_all
    mbase_sk = mbf_sk * mbase_all
    return mbase_cr, mbase_ms, mbase_fat, mbase_sk


def local_q_work(bmr, par):
    """Calculate local thermogenesis by work [W].

    Parameters
    ----------
    bmr : float
        Basal metabolic rate [W].
    par : float
        Physical activity ratio [-].

    Returns
    -------
    q_work : array
        Local thermogenesis by work [W].

    Raises
    ------
    ValueError
        If par is less than 1.

    """
    if par < 1:
        raise ValueError("par must be 1 or more")

    q_work_all = (par - 1) * bmr

    # Distribution coefficient of thermogenesis by work
    workf = np.array(
        [
            0,
            0,
            0.091,
            0.08,
            0.129,
            0.0262,
            0.0139,
            0.005,
            0.0262,
            0.0139,
            0.005,
            0.2010,
            0.0990,
            0.005,
            0.2010,
            0.0990,
            0.005,
        ],
    )
    q_work = q_work_all * workf
    return q_work


PRE_SHIV = 0


def shivering(
    err_cr: np.ndarray,
    err_sk: np.ndarray,
    t_core: np.ndarray,
    t_skin: np.ndarray,
    height,
    weight,
    bsa_equation,
    age,
    sex,
    dtime=60,
    options=None,
):
    """Calculate local thermogenesis by shivering [W].

    Parameters
    ----------
    err_cr, err_sk : array
        Difference between set-point and body temperatures [°C].
    t_core, t_skin : array
        Core and skin temperatures [°C].
    height : float, optional
        Body height [m].
    weight : float, optional
        Body weight [kg].
    bsa_equation : str, optional
        The equation name (str) of bsa calculation. Choose one from BodySurfaceAreaEquations.
    age : float, optional
        age [years].
    sex : str, optional
        Choose male or female.
    dtime : float, optional
        Interval of analysis time.

    Returns
    -------
    q_shiv : array
        Local thermogenesis by shivering [W].

    """
    # Integrated error signal in the warm and cold receptors
    wrms, clds = error_signals(err_sk)

    # Distribution coefficient of thermogenesis by shivering
    shivf = np.array(
        [
            0.0339,
            0.0436,
            0.27394,
            0.24102,
            0.38754,
            0.00243,
            0.00137,
            0.0002,
            0.00243,
            0.00137,
            0.0002,
            0.0039,
            0.00175,
            0.00035,
            0.0039,
            0.00175,
            0.00035,
        ],
    )
    # integrated error signal of shivering
    sig_shiv = 24.36 * clds * (-err_cr[0])
    sig_shiv = max(sig_shiv, 0)

    if options:
        if options["shivering_threshold"]:
            # Asaka, 2016
            # Threshold of starting shivering
            tskm = np.average(t_skin, weights=cons.Default.local_bsa)  # Mean skin temp.
            if tskm < 31:
                thres = 36.6
            elif sex == Sex.male.value:
                thres = -0.2436 * tskm + 44.10
            else:  # sex == "female":
                thres = -0.2250 * tskm + 43.05
            # Second threshold of starting shivering
            if thres < t_core[0]:
                sig_shiv = 0

    # TODO I do not think this is well written code
    global PRE_SHIV  # Previous shivering thermogenesis [W]
    if options:
        if options["limit_dshiv/dt"]:
            dshiv = sig_shiv - PRE_SHIV  # Asaka, 2016 dshiv < 0.0077 [W/s]
            if options["limit_dshiv/dt"] is True:  # default is 0.0077 [W/s]
                limit_dshiv = 0.0077 * dtime
            else:
                limit_dshiv = options["limit_dshiv/dt"] * dtime
            if dshiv > limit_dshiv:
                sig_shiv = limit_dshiv + PRE_SHIV
            elif dshiv < -limit_dshiv:
                sig_shiv = -limit_dshiv + PRE_SHIV
        PRE_SHIV = sig_shiv

    # Signal sd_shiv by aging
    if age < 30:
        sd_shiv = np.ones(Default.num_body_parts)
    elif age < 40:
        sd_shiv = np.ones(Default.num_body_parts) * 0.97514
    elif age < 50:
        sd_shiv = np.ones(Default.num_body_parts) * 0.95028
    elif age < 60:
        sd_shiv = np.ones(Default.num_body_parts) * 0.92818
    elif age < 70:
        sd_shiv = np.ones(Default.num_body_parts) * 0.90055
    elif age < 80:
        sd_shiv = np.ones(Default.num_body_parts) * 0.86188
    else:  # age >= 80
        sd_shiv = np.ones(Default.num_body_parts) * 0.82597

    # Ratio of body surface area to the standard body [-]
    bsar = cons.bsa_rate(height, weight, bsa_equation)

    # Local thermogenesis by shivering [W]
    q_shiv = shivf * bsar * sd_shiv * sig_shiv
    return q_shiv


def nonshivering(
    err_sk,
    height,
    weight,
    bsa_equation,
    age,
    cold_acclimation=False,
    batpositive=True,
):
    """Calculate local metabolic rate by non-shivering [W].

    Parameters
    ----------
    err_sk : array
        Difference between set-point and body temperatures [°C].
    height : float, optional
        Body height [m].
    weight : float, optional
        Body weight [kg].
    bsa_equation : str, optional
        The equation name (str) of bsa calculation. Choose one from BodySurfaceAreaEquations.
    age : float, optional
        age [years].
    cold_acclimation : bool, optional
        Whether the subject acclimates cold environment or not.
        The default is False.
    batpositive : bool, optional
        Whether BAT activity is positive or not.
        The default is True.

    Returns
    -------
    q_nst : array
        Local metabolic rate by non-shivering [W].

    """
    # NST (Non-Shivering Thermogenesis) model, Asaka, 2016
    wrms, clds = error_signals(err_sk)

    # BMI (Body Mass Index)
    bmi = weight / height**2

    # BAT: brown adipose tissue [SUV]
    bat = 10 ** (-0.10502 * bmi + 2.7708)

    # age factor
    if age < 30:
        bat *= 1.61
    elif age < 40:
        bat *= 1.00
    else:  # age >= 40
        bat *= 0.80

    if cold_acclimation:
        bat += 3.46

    if not batpositive:
        # incidence age factor: T.Yoneshiro 2011
        if age < 30:  # age = 20s or younger
            bat *= 44 / 83
        elif age < 40:  # age = 30s
            bat *= 15 / 38
        elif age < 50:  # age = 40s
            bat *= 7 / 26
        elif age < 60:  # age = 50s
            bat *= 1 / 8
        else:  # age > 60
            bat *= 0

    # NST limit
    thres = (1.80 * bat + 2.43) + 5.62  # [W]

    sig_nst = 2.8 * clds  # [W]
    sig_nst = min(sig_nst, thres)

    # Distribution coefficient of thermogenesis by non-shivering
    nstf = np.array(
        [
            0.000,
            0.190,
            0.000,
            0.190,
            0.190,
            0.215,
            0.000,
            0.000,
            0.215,
            0.000,
            0.000,
            0.000,
            0.000,
            0.000,
            0.000,
            0.000,
            0.000,
        ],
    )

    # Ratio of body surface area to the standard body [-]
    bsar = cons.bsa_rate(height, weight, bsa_equation)

    # Local thermogenesis by non-shivering [W]
    q_nst = bsar * nstf * sig_nst
    return q_nst


def sum_m(mbase: np.ndarray, q_work, q_shiv, q_nst):
    """Calculate total thermogenesis in each layer [W].

    Parameters
    ----------
    mbase : array
        Local basal metabolic rate (Mbase) [W].
    q_work : array
        Local thermogenesis by work [W].
    q_shiv : array
        Local thermogenesis by shivering [W].
    q_nst : array
        Local thermogenesis by non-shivering [W].

    Returns
    -------
    q_thermogenesis_core, q_thermogenesis_muscle, q_thermogenesis_fat, q_thermogenesis_skin : array
        Total thermogenesis in core, muscle, fat, skin layers [W].

    """
    q_thermogenesis_core = mbase[0].copy()
    q_thermogenesis_muscle = mbase[1].copy()
    q_thermogenesis_fat = mbase[2].copy()
    q_thermogenesis_skin = mbase[3].copy()

    for i, bn in enumerate(JOS3BodyParts.get_attribute_names()):
        # If the segment has a muscle layer, muscle thermogenesis increases by the activity.
        if IDICT[bn]["muscle"] is not None:
            q_thermogenesis_muscle[i] += q_work[i] + q_shiv[i]
        # In other segments, core thermogenesis increase, instead of muscle.
        else:
            q_thermogenesis_core[i] += q_work[i] + q_shiv[i]
    q_thermogenesis_core += q_nst  # Non-shivering thermogenesis occurs in core layers
    return (
        q_thermogenesis_core,
        q_thermogenesis_muscle,
        q_thermogenesis_fat,
        q_thermogenesis_skin,
    )


def cr_ms_fat_blood_flow(
    q_work,
    q_shiv,
    height,
    weight,
    bsa_equation,
    age,
    ci,
):
    """Calculate core, muscle and fat blood flow rate [L/h].

    Parameters
    ----------
    q_work : array
        Heat production by work [W].
    q_shiv : array
        Heat production by shivering [W].
    height : float, optional
        Body height [m].
    weight : float, optional
        Body weight [kg].
    bsa_equation : str, optional
        The equation name (str) of bsa calculation. Choose one from BodySurfaceAreaEquations.
    age : float, optional
        age [years].
    ci : float, optional
        Cardiac index [L/min/㎡].

    Returns
    -------
    bf_core, bf_muscle, bf_fat : array
        Core, muscle and fat blood flow rate [L/h].

    """
    # Basal blood flow rate [L/h]
    # core, CBFB
    bfb_core = np.array(
        [
            35.251,
            15.240,
            89.214,
            87.663,
            18.686,
            1.808,
            0.940,
            0.217,
            1.808,
            0.940,
            0.217,
            1.406,
            0.164,
            0.080,
            1.406,
            0.164,
            0.080,
        ],
    )
    # muscle, MSBFB
    bfb_muscle = np.array(
        [
            0.682,
            0.0,
            0.0,
            0.0,
            12.614,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
    )
    # fat, FTBFB
    bfb_fat = np.array(
        [
            0.265,
            0.0,
            0.0,
            0.0,
            2.219,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
    )

    bfb_rate = cons.bfb_rate(height, weight, bsa_equation, age, ci)
    bf_core = bfb_core * bfb_rate
    bf_muscle = bfb_muscle * bfb_rate
    bf_fat = bfb_fat * bfb_rate

    for i, bn in enumerate(JOS3BodyParts.get_attribute_names()):
        # If the segment has a muscle layer, muscle blood flow increases.
        if IDICT[bn]["muscle"] is not None:
            bf_muscle[i] += (q_work[i] + q_shiv[i]) / 1.163
        # In other segments, core blood flow increase, instead of muscle blood flow.
        else:
            bf_core[i] += (q_work[i] + q_shiv[i]) / 1.163
    return bf_core, bf_muscle, bf_fat


def sum_bf(bf_core, bf_muscle, bf_fat, bf_skin, bf_ava_hand, bf_ava_foot):
    """Sum the total blood flow in various body parts.

    Parameters
    ----------
    bf_core : array
        Blood flow rate in the core region [L/h].
    bf_muscle : array
        Blood flow rate in the muscle region [L/h].
    bf_fat : array
        Blood flow rate in the fat region [L/h].
    bf_skin : array
        Blood flow rate in the skin region [L/h].
    bf_ava_hand : array
        AVA blood flow rate in one hand [L/h].
    bf_ava_foot : array
        AVA blood flow rate in one foot [L/h].

    Returns
    -------
    co : float
        Cardiac output (the sum of the whole blood flow rate) [L/h].

    """
    # Cardiac output (CO)
    co = 0
    co += bf_core.sum()
    co += bf_muscle.sum()
    co += bf_fat.sum()
    co += bf_skin.sum()
    co += 2 * bf_ava_hand
    co += 2 * bf_ava_foot
    return co


def resp_heat_loss(tdb: float, p_a: float, q_thermogenesis_total: float):
    """Calculate heat loss by respiration [W].

    Parameters
    ----------
    tdb : float
        Dry bulb air temperature [°C].
    p_a : float
        Water vapor pressure in the ambient air [kPa].
    q_thermogenesis_total : float
        Total thermogenesis [W].

    Returns
    -------
    res_sh : float
        Sensible heat loss by respiration [W].
    res_lh : float
        Latent heat loss by respiration [W].

    """
    res_sh = 0.0014 * q_thermogenesis_total * (34 - tdb)  # Sensible heat loss
    res_lh = 0.0173 * q_thermogenesis_total * (5.87 - p_a)  # Latent heat loss
    return res_sh, res_lh



import time

import numpy as np

from pythermalcomfort.models import utci

utci_val = utci(tdb=29, tr=30, v=1, rh=60)
print(utci_val)

utci_val = utci(tdb=60, tr=30, v=1, rh=60)
print(utci_val)
print(utci(tdb=60, tr=30, v=1, rh=60, limit_inputs=False))

utci_val = utci(tdb=77, tr=77, v=6.56168, rh=60, units="IP")
print(utci_val)

# numpy examples
utci_val = utci(
    tdb=[29, 29, 25],
    tr=[30, 30, 25],
    v=[1, 2, 1],
    rh=[60, 60, 50],
)
print(utci_val.utci)

utci_val = utci(
    tdb=[29, 29, 25],
    tr=[30, 30, 25],
    v=[1, 2, 1],
    rh=[60, 60, 50],
)
print(utci_val)
print(utci_val["utci"])
print(utci_val["stress_category"])

iterations = 100000
tdb = np.empty(iterations)
tdb.fill(25)
start = time.time()
utci_val = utci(tdb=tdb, tr=30, v=1, rh=60)
end = time.time()
print(end - start)



import numpy as np


def valid_range(x, valid) -> np.ndarray:
    """Filter values based on a valid range."""
    return np.where((x >= valid[0]) & (x <= valid[1]), x, np.nan)


def mapping(value, map_dictionary, right=True):
    """Map a temperature array to stress categories.

    Parameters
    ----------
    value : float, array-like
        Temperature to map.
    map_dictionary: dict
        Dictionary used to map the values
    right: bool, optional
        Indicating whether the intervals include the right or the left bin edge.

    Returns
    -------
    Stress category for each input temperature.

    """
    bins = np.array(list(map_dictionary.keys()))
    words = np.append(np.array(list(map_dictionary.values())), "unknown")
    return words[np.digitize(value, bins, right=right)]



from __future__ import annotations

from typing import Literal

import numpy as np

from pythermalcomfort.classes_input import CloTOutInputs
from pythermalcomfort.classes_return import CloTOut
from pythermalcomfort.utilities import Units, units_converter


def clo_tout(
    tout: float | list[float],
    units: Literal["SI", "IP"] = Units.SI.value,
) -> CloTOut:
    """Calculate representative clothing insulation Icl based on outdoor air
    temperature at 06:00 a.m. [Schiavon2013]_.

    Parameters
    ----------
    tout : float or list of floats
        Outdoor air temperature at 06:00 a.m., default in [°C] or [°F] if `units` = 'IP'.
    units : str, optional
        Select the SI (International System of Units) or the IP (Imperial Units) system.
        Supported values are 'SI' and 'IP'. Defaults to 'SI'.

    Returns
    -------
    CloTOut
        A dataclass containing the representative clothing insulation Icl. See :py:class:`~pythermalcomfort.classes_return.CloTOut` for more details.
        To access the `clo_tout` value, use the `clo_tout` attribute of the returned `CloTOut` instance, e.g., `result.clo_tout`.

    Raises
    ------
    TypeError
        If `tout` is not a float, int, NumPy array, or a list of floats or integers.
    ValueError
        If an invalid unit is provided or non-numeric elements are found in `tout`.

    Notes
    -----
    .. note::
        The ASHRAE 55 2020 states that it is acceptable to determine the clothing
        insulation Icl using this equation in mechanically conditioned buildings [55ASHRAE2023]_.

    .. note::
        Limitations:
        - This equation may not be accurate for extreme temperature ranges.

    Examples
    --------
    .. code-block:: python

        from pythermalcomfort.models import clo_tout

        result = clo_tout(tout=27)
        print(result.clo_tout)  # 0.46

        result = clo_tout(tout=[27, 25])
        print(result.clo_tout)  # array([0.46, 0.47])

    """
    # Validate inputs using the CloTOutInputs class
    CloTOutInputs(
        tout=tout,
        units=units,
    )

    # Convert tout to NumPy array for vectorized operations
    tout = np.array(tout)

    # Convert units if necessary
    if units.upper() == Units.IP.value:
        tout = units_converter(tmp=tout)[0]

    clo = np.where(tout < 26, np.power(10, -0.1635 - 0.0066 * tout), 0.46)
    clo = np.where(tout < 5, 0.818 - 0.0364 * tout, clo)
    clo = np.where(tout < -5, 1, clo)

    return CloTOut(clo_tout=np.around(clo, 2))



from __future__ import annotations

import numpy as np

from pythermalcomfort.classes_input import WCIInputs
from pythermalcomfort.classes_return import WCI


def wci(
    tdb: float | list[float],
    v: float | list[float],
    round_output: bool = True,
) -> WCI:
    """Calculate the Wind Chill Index (WCI) in accordance with the ASHRAE 2017 Handbook Fundamentals - Chapter 9 [ashrae2017]_.

    The wind chill index (WCI) is an empirical index based on cooling measurements
    taken on a cylindrical flask partially filled with water in Antarctica
    (Siple and Passel 1945). For a surface temperature of 33°C, the index describes
    the rate of heat loss from the cylinder via radiation and convection as a function
    of ambient temperature and wind velocity.

    This formulation has been met with some valid criticism. WCI is unlikely to be an
    accurate measure of heat loss from exposed flesh, which differs from plastic in terms
    of curvature, roughness, and radiation exchange qualities, and is always below 33°C
    in a cold environment. Furthermore, the equation's values peak at 90 km/h and then
    decline as velocity increases. Nonetheless, this score reliably represents the
    combined effects of temperature and wind on subjective discomfort for velocities
    below 80 km/h [ashrae2017]_.

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, [°C].
    v : float or list of floats
        Wind speed 10m above ground level, [m/s].
    round_output : bool, optional
        If True, rounds output value. If False, it does not round it. Defaults to True.

    Returns
    -------
    WCI
        A dataclass containing the Wind Chill Index. See :py:class:`~pythermalcomfort.classes_return.WCI` for more details.
        To access the `wci` value, use the `wci` attribute of the returned `WCI` instance, e.g., `result.wci`.

    Examples
    --------
    .. code-block:: python

        from pythermalcomfort.models import wci

        result = wci(tdb=-5, v=5.5)
        print(result.wci)  # 1255.2

        result = wci(tdb=[-5, -10], v=[5.5, 10], round_output=True)
        print(result.wci)  # [1255.2 1603.9]

    """
    # Validate inputs using the WCYInputs class
    WCIInputs(
        tdb=tdb,
        v=v,
        round_output=round_output,
    )

    tdb = np.array(tdb)
    v = np.array(v)

    _wci = (10.45 + 10 * v**0.5 - v) * (33 - tdb)

    # the factor 1.163 is used to convert to W/m^2
    _wci = _wci * 1.163

    if round_output:
        _wci = np.around(_wci, 1)

    return WCI(wci=_wci)



from pythermalcomfort.models import clo_tout
from pythermalcomfort.utilities import Units
from tests.conftest import Urls, retrieve_reference_table, validate_result


def test_clo_tout(get_test_url, retrieve_data) -> None:
    """Test that the function calculates the clothing thermal insulation (Clo) for various inputs."""
    reference_table = retrieve_reference_table(
        get_test_url,
        retrieve_data,
        Urls.CLO_TOUT.name,
    )
    tolerance = reference_table["tolerance"]

    for entry in reference_table["data"]:
        inputs = entry["inputs"]
        outputs = entry["outputs"]
        result = clo_tout(
            tout=inputs["tout"],
            units=inputs.get("units", Units.SI.value),
        )

        validate_result(result, outputs, tolerance)



from __future__ import annotations

import numpy as np

from pythermalcomfort.classes_input import SETInputs
from pythermalcomfort.classes_return import SET
from pythermalcomfort.models.two_nodes_gagge import two_nodes_gagge
from pythermalcomfort.utilities import (
    Models,
    Postures,
    _check_standard_compliance_array,
)


def set_tmp(
    tdb: float | list[float],
    tr: float | list[float],
    v: float | list[float],
    rh: float | list[float],
    met: float | list[float],
    clo: float | list[float],
    wme: float | list[float] = 0.0,
    body_surface_area: float | list[float] = 1.8258,
    p_atm: float | list[float] = 101325,
    position: str = Postures.standing.value,
    limit_inputs: bool = True,
    round_output: bool = True,
    calculate_ce: bool = False,
) -> SET:
    """Calculate the Standard Effective Temperature (SET).

    The SET is the
    temperature of a hypothetical isothermal environment at 50% (rh), <0.1 m/s
    (20 fpm) average air speed (v), and tr = tdb, in which the total heat loss
    from the skin of an imaginary occupant wearing clothing, standardized for
    the activity concerned is the same as that from a person in the actual
    environment with actual clothing and activity level. [Gagge1986]_

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, [°C].
    tr : float or list of floats
        Mean radiant temperature, [°C].
    v : float or list of floats
        Air speed, [m/s].
    rh : float or list of floats
        Relative humidity, [%].
    met : float or list of floats
        Metabolic rate, [met].
    clo : float or list of floats
        Clothing insulation, [clo].
    wme : float or list of floats, optional
        External work, [met]. Defaults to 0.
    body_surface_area : float or list of floats, optional
        Body surface area, default value 1.8258 [m2]

        .. note::
            The body surface area can be calculated using the function
            :py:meth:`pythermalcomfort.utilities.body_surface_area`.

    p_atm : float or list of floats, optional
        Atmospheric pressure, default value 101325 [Pa]
    position : str, optional
        Select either "sitting" or "standing". Defaults to "standing".
    limit_inputs : bool, optional
        If True, limits the inputs to the standard applicability limits. Defaults to True.

        .. note::
            By default, if the inputs are outside the standard applicability limits the
            function returns nan. If False returns values regardless of the input values.
            The limits are 10 < tdb [°C] < 40, 10 < tr [°C] < 40,
            0 < v [m/s] < 2, 1 < met [met] < 4, and 0 < clo [clo] < 1.5.
    round_output : bool, optional
        If True, rounds output value. If False, it does not round it. Defaults to True.

    Returns
    -------
    SET
        A dataclass containing the Standard Effective Temperature. See :py:class:`~pythermalcomfort.classes_return.SetTmp` for more details.
        To access the `set` value, use the corresponding attribute of the returned `SetTmp` instance, e.g., `result.set`.

    Examples
    --------
    .. code-block:: python

        from pythermalcomfort.models import set_tmp

        result = set_tmp(tdb=25, tr=25, v=0.1, rh=50, met=1.2, clo=0.5)
        print(result.set)  # 24.3

        result = set_tmp(tdb=[25, 25], tr=25, v=0.1, rh=50, met=1.2, clo=0.5)
        print(result.set)  # [24.3, 24.3]

    """
    tdb = np.array(tdb)
    tr = np.array(tr)
    v = np.array(v)
    rh = np.array(rh)
    met = np.array(met)
    clo = np.array(clo)
    wme = np.array(wme)

    # Validate inputs using the SetTmpInputs class
    SETInputs(
        tdb=tdb,
        tr=tr,
        v=v,
        rh=rh,
        met=met,
        clo=clo,
        wme=wme,
        body_surface_area=body_surface_area,
        p_atm=p_atm,
        position=position,
        limit_inputs=limit_inputs,
    )

    set_array = two_nodes_gagge(
        tdb=tdb,
        tr=tr,
        v=v,
        rh=rh,
        met=met,
        clo=clo,
        wme=wme,
        body_surface_area=body_surface_area,
        p_atm=p_atm,
        position=position,
        calculate_ce=calculate_ce,
        round_output=False,
    ).set

    if limit_inputs:
        (
            tdb_valid,
            tr_valid,
            v_valid,
            met_valid,
            clo_valid,
        ) = _check_standard_compliance_array(
            standard=Models.ashrae_55_2023.value,
            tdb=tdb,
            tr=tr,
            v=v,
            met=met,
            clo=clo,
        )
        all_valid = ~(
            np.isnan(tdb_valid)
            | np.isnan(tr_valid)
            | np.isnan(v_valid)
            | np.isnan(met_valid)
            | np.isnan(clo_valid)
        )
        set_array = np.where(all_valid, set_array, np.nan)

    if round_output:
        set_array = np.around(set_array, 1)

    return SET(set=set_array)



from __future__ import annotations

import numpy as np
import numpy.typing as npt
from scipy import optimize

from pythermalcomfort.classes_input import PETSteadyInputs
from pythermalcomfort.classes_return import PETSteady
from pythermalcomfort.utilities import Postures, Sex, body_surface_area, p_sat


def pet_steady(
    tdb: float | list[float],
    tr: float | list[float],
    v: float | list[float],
    rh: float | list[float],
    met: float | list[float],
    clo: float | list[float],
    p_atm: float | list[float] = 1013.25,
    position: str | list[str] = Postures.sitting.value,
    age: int | list[int] = 23,
    sex: int | list[int] = Sex.male.value,
    weight: float | list[float] = 75,
    height: float | list[float] = 1.8,
    wme: float | list[float] = 0,
) -> PETSteady:
    """Calculate the steady physiological equivalent temperature (PET) using the Munich
    Energy-balance Model for Individuals (MEMI) to simulate the human body's thermal
    state in a medically realistic manner.

    PET is defined as the air temperature
    at which, in a typical indoor setting the heat budget of the human body is balanced
    with the same core and skin temperature as under the complex outdoor conditions to be
    assessed [Hoppe1999]_.
    The following assumptions are made for the indoor reference climate: tdb = tr, v = 0.1
    m/s, water vapour pressure = 12 hPa, clo = 0.9 clo, and met = 1.37 met + basic
    metabolism.
    PET allows a layperson to compare the total effects of complex thermal circumstances
    outside with his or her own personal experience indoors in this way. This function
    solves the heat balances without accounting for heat storage in the human body.

    The PET was originally proposed by Hoppe [Hoppe1999]_. In 2018, Walther and Goestchel [Walther2018]_
    proposed a correction of the original model, purging the errors in the
    PET calculation routine, and implementing a state-of-the-art vapour diffusion model.
    Walther and Goestchel (2018) model is therefore used to calculate the PET.

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, [°C].
    tr : float or list of floats
        Mean radiant temperature, [°C].
    v : float or list of floats
        Wind speed, [m/s].
    rh : float or list of floats
        Relative humidity, [%].
    met : float or list of floats
        Metabolic rate, [met].
    clo : float or list of floats
        Clothing insulation, [clo].
    p_atm : float or list of floats, optional
        Atmospheric pressure, [hPa]. Defaults to 1013.25.
    position : str or list of str, optional
        Position of the person "sitting", "standing", "standing, forced convection". Defaults to "sitting".
    age : int or list of ints, optional
        Age of the person. Defaults to 23.
    sex : str or list of str, optional
        Sex of the person "male" or "female". Defaults to "male"
    weight : float or list of floats, optional
        Weight of the person, [kg]. Defaults to 75.
    height : float or list of floats, optional
        Height of the person, [m]. Defaults to 1.8.
    wme : float or list of floats, optional
        External work, [met]. Defaults to 0.

    Returns
    -------
    PETSteady
        A dataclass containing the Physiological Equivalent Temperature. See :py:class:`~pythermalcomfort.classes_return.PETSteady` for more details.
        To access the `pet` value, use the `pet` attribute of the returned `PETSteady` instance, e.g., `result.pet`.

    Examples
    --------
    .. code-block:: python

        from pythermalcomfort.models import pet_steady

        result = pet_steady(tdb=25, tr=25, v=0.1, rh=50, met=1.2, clo=0.5)
        print(result.pet)  # 24.67

        result = pet_steady(
            tdb=[25, 30],
            tr=[25, 30],
            v=[0.1, 0.2],
            rh=[50, 60],
            met=[1.2, 1.4],
            clo=[0.5, 0.6],
        )
        print(result.pet)  # [24.67 31.46]

    """
    # Validate inputs using the PETSteadyInputs class
    PETSteadyInputs(
        tdb=tdb,
        tr=tr,
        v=v,
        rh=rh,
        met=met,
        clo=clo,
        p_atm=p_atm,
        position=position,
        age=age,
        sex=sex,
        weight=weight,
        height=height,
        wme=wme,
    )

    return PETSteady(
        pet=_pet_steady_vectorised(
            tdb=tdb,
            tr=tr,
            v=v,
            rh=rh,
            met=met,
            clo=clo,
            p_atm=p_atm,
            position=position,
            age=age,
            sex=sex,
            weight=weight,
            height=height,
            wme=wme,
        ),
    )


@np.vectorize
def _pet_steady_vectorised(
    tdb,
    tr,
    v,
    rh,
    met,
    clo,
    p_atm,
    position,
    age,
    sex,
    weight,
    height,
    wme,
) -> npt.ArrayLike:
    met_factor = 58.2  # met conversion factor
    met = met * met_factor  # metabolic rate

    def vasomotricity(t_cr, t_sk):
        """Vasomotricity (blood flow) in function of the core and skin temperatures.

        Parameters
        ----------
        t_cr : float
            The body core temperature, [°C]
        t_sk : float
            The body skin temperature, [°C]

        Returns
        -------
        dict
            "m_blood": Blood flow rate, [kg/m2/h] and "alpha": repartition of body
            mass
            between core and skin [].

        """
        # skin and core temperatures set values
        tc_set = 36.6  # 36.8
        tsk_set = 34  # 33.7
        # Set value signals
        sig_skin = tsk_set - t_sk
        sig_core = t_cr - tc_set
        if sig_core < 0:
            # In this case, T_core<Tc_set --> the blood flow is reduced
            sig_core = 0.0
        if sig_skin < 0:
            # In this case, Tsk>Tsk_set --> the blood flow is increased
            sig_skin = 0.0
        # 6.3 L/m^2/h is the set value of the blood flow
        m_blood = (6.3 + 75.0 * sig_core) / (1.0 + 0.5 * sig_skin)
        # 90 L/m^2/h is the blood flow upper limit
        if m_blood > 90:
            m_blood = 90.0
        # in other models, alpha is used to update tbody
        alpha = 0.0417737 + 0.7451833 / (m_blood + 0.585417)

        return {"m_blood": m_blood, "alpha": alpha}

    def sweat_rate(t_body):
        """Sweating mechanism depending on the body and core temperatures.

        Parameters
        ----------
        t_body : float
            weighted average between skin and core temperatures, [°C]

        Returns
        -------
        m_rsw : float
            The sweating flow rate, [g/m2/h].

        """
        tc_set = 36.6  # 36.8
        tsk_set = 34  # 33.7
        tbody_set = 0.1 * tsk_set + 0.9 * tc_set  # Calculation of the body
        # temperature
        # through a weighted average
        sig_body = t_body - tbody_set
        if sig_body < 0:
            # In this case, Tbody<Tbody_set --> The sweat flow is 0
            sig_body = 0.0
        # from Gagge's model
        m_rsw = 304.94 * sig_body
        # 500 g/m^2/h is the upper sweat rate limit
        m_rsw = min(m_rsw, 500)

        return m_rsw

    def solve_pet(
        t_arr,
        _tdb,
        _tr,
        _v=0.1,
        _rh=50,
        _met=80,
        _clo=0.9,
        actual_environment=False,
    ):
        """Solve PET by either computing the vectorial balance of the three unknown temperatures
        (T_core, T_sk, T_clo) or by finding the environment operative temperature that yields
        the same energy balance as the actual environment.

        Parameters
        ----------
        t_arr : list or list of floats
            [T_core, T_sk, T_clo], [°C]
        _tdb : float
            dry bulb air temperature, [°C]
        _tr : float
            mean radiant temperature, [°C]
        _v : float, default 0.1 m/s for the reference environment
            air speed, [m/s]
        _rh : float, default 50 % for the reference environment
            relative humidity, [%]
        _met : float, default 80 W for the reference environment
            metabolic rate, [W/m2]
        _clo : float, default 0.9 clo for the reference environment
            clothing insulation, [clo]
        actual_environment : boolean
            True=solve 3eqs/3unknowns, False=solve for PET

        Returns
        -------
        float or list
            PET or energy balance.

        """
        e_skin = 0.99  # Skin emissivity
        e_clo = 0.95  # Clothing emissivity
        h_vap = 2.42 * 10**6  # Latent heat of evaporation [J/Kg]
        sbc = 5.67 * 10**-8  # Stefan-Boltzmann constant [W/(m2*K^(-4))]
        cb = 3640  # Blood specific heat [J/kg/k]

        t_cr, t_sk, t_clo = t_arr
        e_bal_vec = [
            0.0,
            0.0,
            0.0,
        ]  # required for the vectorial expression of the balance
        # Area parameters of the body:
        a_dubois = body_surface_area(weight, height)
        # Base metabolism for men and women in [W]
        met_female = (
            3.19
            * weight**0.75
            * (
                1.0
                + 0.004 * (30.0 - age)
                + 0.018 * (height * 100.0 / weight ** (1.0 / 3.0) - 42.1)
            )
        )
        met_male = (
            3.45
            * weight**0.75
            * (
                1.0
                + 0.004 * (30.0 - age)
                + 0.01 * (height * 100.0 / weight ** (1.0 / 3.0) - 43.4)
            )
        )
        # Attribution of internal energy depending on the sex of the subject
        met_correction = met_male if sex == Sex.male.value else met_female

        # Source term : metabolic activity
        he = (_met + met_correction) / a_dubois
        # impact of efficiency
        h = he * (1.0 - wme)  # [W/m2]

        # correction for wind
        i_m = 0.38  # Woodcock ratio for vapour transfer through clothing [-]

        # Calculation of the Burton surface increase coefficient, k = 0.31 for Hoeppe:
        fcl = (
            1 + 0.31 * _clo
        )  # Increase heat exchange surface depending on clothing level
        f_a_cl = (173.51 * _clo - 2.36 - 100.76 * _clo * _clo + 19.28 * _clo**3.0) / 100
        a_clo = a_dubois * f_a_cl + a_dubois * (fcl - 1.0)  # clothed body surface area

        f_eff = (
            0.696 if position == Postures.standing.value else 0.725
        )  # effective radiation factor

        a_r_eff = (
            a_dubois * f_eff
        )  # Effective radiative area depending on the position of the subject

        # Partial pressure of water in the air
        vpa = _rh / 100.0 * p_sat(_tdb) / 100  # [hPa]
        if not actual_environment:  # mode=False means we are calculating the PET
            vpa = 12  # [hPa] vapour pressure of the standard environment

        # Convection coefficient depending on wind velocity and subject position
        hc = 2.67 + 6.5 * _v**0.67  # sitting
        if position == Postures.standing.value:  # standing
            hc = 2.26 + 7.42 * _v**0.67
        if position == "standing, forced convection":  # standing, forced convection
            hc = 8.6 * _v**0.513
        # h_cc corrected convective heat transfer coefficient
        h_cc = 3.0 * pow(p_atm / 1013.25, 0.53)
        hc = max(h_cc, hc)
        # modification of hc with the total pressure
        hc = hc * (p_atm / 1013.25) ** 0.55

        # Respiratory energy losses
        t_exp = 0.47 * _tdb + 21.0  # Expired air temperature calculation [degC]
        d_vent_pulm = he * 1.44 * 10.0 ** (-6.0)  # breathing flow rate
        c_res = 1010 * (_tdb - t_exp) * d_vent_pulm  # Sensible heat energy loss [W/m2]
        vpexp = p_sat(t_exp) / 100  # Latent heat energy loss [hPa]
        q_res = 0.623 * h_vap / p_atm * (vpa - vpexp) * d_vent_pulm  # [W/m2]
        ere = c_res + q_res  # [W/m2]

        # Calculation of the equivalent thermal resistance of body tissues
        alpha = vasomotricity(t_cr, t_sk)["alpha"]
        tbody = alpha * t_sk + (1 - alpha) * t_cr

        # Clothed fraction of the body approximation
        r_cl = _clo / 6.45  # Conversion in [m2.K/W]
        y = 0
        f_a_cl = min(f_a_cl, 1.0)
        if _clo >= 2.0:
            y = 1.0
        if 0.6 < _clo < 2.0:
            y = (height - 0.2) / height
        if 0.6 >= _clo > 0.3:
            y = 0.5
        if 0.3 >= _clo > 0.0:
            y = 0.1
        # calculation of the clothing radius depending on the clothing level (6.28 = 2*
        # pi !)
        r2 = a_dubois * (fcl - 1.0 + f_a_cl) / (6.28 * height * y)  # External radius
        r1 = f_a_cl * a_dubois / (6.28 * height * y)  # Internal radius
        di = r2 - r1
        # Calculation of the equivalent thermal resistance of body tissues
        htcl = 6.28 * height * y * di / (r_cl * np.log(r2 / r1) * a_clo)  # [W/(m2.K)]

        # Calculation of sweat losses
        qmsw = sweat_rate(tbody)
        # h_vap/1000 = 2400 000[J/kg] divided by 1000 = [J/g] // qwsw/3600 for [g/m2/h]
        # to [
        # g/m2/s]
        esw = h_vap / 1000 * qmsw / 3600  # [W/m2]
        # Saturation vapor pressure at temperature Tsk
        p_v_sk = p_sat(t_sk) / 100  # hPa
        # Calculation of vapour transfer
        lr = 16.7 * 10 ** (-1)  # [K/hPa] Lewis ratio
        he_diff = hc * lr  # diffusion coefficient of air layer
        fecl = 1 / (1 + 0.92 * hc * r_cl)  # Burton efficiency factor
        e_max = he_diff * fecl * (p_v_sk - vpa)  # maximum diffusion at skin surface
        if e_max == 0:  # added this otherwise e_req / e_max cannot be calculated
            e_max = 0.001
        w = esw / e_max  # skin wettedness
        if w > 1:
            w = 1
            delta = esw - e_max
            if delta < 0:
                esw = e_max
        esw = max(esw, 0)
        # i_m= Woodcock's ratio (see above)
        r_ecl = (1 / (fcl * hc) + r_cl) / (
            lr * i_m
        )  # clothing vapour transfer resistance after Woodcock's method
        ediff = (1 - w) * (p_v_sk - vpa) / r_ecl  # diffusion heat transfer
        evap = -(ediff + esw)  # [W/m2]

        # Radiation losses bare skin
        r_bare = (
            a_r_eff
            * (1.0 - f_a_cl)
            * e_skin
            * sbc
            * ((_tr + 273.15) ** 4.0 - (t_sk + 273.15) ** 4.0)
            / a_dubois
        )
        # ... for clothed area
        r_clo = (
            f_eff
            * a_clo
            * e_clo
            * sbc
            * ((_tr + 273.15) ** 4.0 - (t_clo + 273.15) ** 4.0)
            / a_dubois
        )
        r_sum = r_clo + r_bare  # radiation total

        # Convection losses for bare skin
        c_bare = hc * (_tdb - t_sk) * a_dubois * (1.0 - f_a_cl) / a_dubois  # [W/m^2]
        # ... for clothed area
        c_clo = hc * (_tdb - t_clo) * a_clo / a_dubois  # [W/m^2]
        csum = c_clo + c_bare  # convection total

        # Balance equations of the 3-nodes model
        e_bal_vec[0] = (
            h
            + ere
            - (vasomotricity(t_cr, t_sk)["m_blood"] / 3600 * cb + 5.28) * (t_cr - t_sk)
        )  # Core balance [W/m^2]
        e_bal_vec[1] = (
            r_bare
            + c_bare
            + evap
            + (vasomotricity(t_cr, t_sk)["m_blood"] / 3600 * cb + 5.28) * (t_cr - t_sk)
            - htcl * (t_sk - t_clo)
        )  # Skin balance [W/m^2]
        e_bal_vec[2] = c_clo + r_clo + htcl * (t_sk - t_clo)  # Clothes balance [W/m^2]
        e_bal_scal = h + ere + r_sum + csum + evap

        # returning either the calculated core, skin, clo temperatures or the PET
        if actual_environment:
            # if we solve for the system we need to return 3 temperatures
            return list(e_bal_vec)
        # solving for the PET requires the scalar balance only
        return e_bal_scal

    def pet_fc(_t_stable):
        """Find the solution.

        Parameters
        ----------
        _t_stable : list or list of floats
            3 temperatures obtained from the actual environment (T_core,T_skin,T_clo).

        Returns
        -------
        float
            The PET comfort index.

        """

        # Definition of a function with the input variables of the PET reference situation
        def f(tx):
            return solve_pet(
                _t_stable,
                _tdb=tx,
                _tr=tx,
                actual_environment=False,
            )

        # solving for PET
        pet_guess = _t_stable[2]  # start with the clothing temperature

        return round(optimize.fsolve(f, pet_guess)[0], 2)

    # initial guess
    t_guess = np.array([36.7, 34, 0.5 * (tdb + tr)])
    # solve for Tc, Tsk, Tcl temperatures
    t_stable = optimize.fsolve(
        solve_pet,
        t_guess,
        args=(
            tdb,
            tr,
            v,
            rh,
            met,
            clo,
            True,
        ),
    )
    # compute PET
    return pet_fc(t_stable)



import pytest

from pythermalcomfort.models import wci
from tests.conftest import Urls, retrieve_reference_table, validate_result


def test_calculates_wci(get_test_url, retrieve_data) -> None:
    """Test that the function calculates the wind chill index (WCI) correctly for various inputs."""
    reference_table = retrieve_reference_table(
        get_test_url,
        retrieve_data,
        Urls.WIND_CHILL.name,
    )
    tolerance = reference_table["tolerance"]

    for entry in reference_table["data"]:
        inputs = entry["inputs"]
        outputs = entry["outputs"]
        result = wci(**inputs)

        validate_result(result, outputs, tolerance)


class TestWc:
    """Test cases for the wind chill index (WCI) model."""

    #  Raises TypeError if tdb parameter is not provided
    def test_raises_type_error_tdb_not_provided(self) -> None:
        """Test that the function raises TypeError if tdb is not provided."""
        with pytest.raises(TypeError):
            wci(v=0.1)

    #  Raises TypeError if v parameter is not provided
    def test_raises_type_error_v_not_provided(self) -> None:
        """Test that the function raises TypeError if v is not provided."""
        with pytest.raises(TypeError):
            wci(tdb=0)

    #  Raises TypeError if tdb parameter is not a float
    def test_raises_type_error_tdb_not_float(self) -> None:
        """Test that the function raises TypeError if tdb is not a float."""
        with pytest.raises(TypeError):
            wci(tdb="0", v=0.1)



import math

import pytest

from pythermalcomfort.models import humidex
from tests.conftest import Urls, retrieve_reference_table, validate_result


def test_humidex(get_test_url, retrieve_data) -> None:
    """Test that the function calculates the Humidex correctly for various inputs."""
    reference_table = retrieve_reference_table(
        get_test_url,
        retrieve_data,
        Urls.HUMIDEX.name,
    )
    tolerance = reference_table["tolerance"]

    for entry in reference_table["data"]:
        inputs = entry["inputs"]
        outputs = entry["outputs"]
        result = humidex(**inputs)

        validate_result(result, outputs, tolerance)

    with pytest.raises(TypeError):
        humidex("25", 50)

    with pytest.raises(TypeError):
        humidex(25, "50")

    with pytest.raises(ValueError):
        humidex(tdb=25, rh=110)

    with pytest.raises(ValueError):
        humidex(25, -10)


def test_humidex_masterson() -> None:
    """Test the Masterson model for Humidex calculations."""
    # TODO move this to shared test
    # I got these values from
    # https://publications.gc.ca/collections/collection_2018/eccc/En57-23-1-79-eng.pdf
    result = humidex(tdb=21, rh=100, model="masterson")
    assert math.isclose(result.humidex, 29.3, abs_tol=0.01)
    assert result.discomfort == "Little or no discomfort"

    result = humidex(tdb=34, rh=100, model="masterson")
    assert math.isclose(result.humidex, 58.4, abs_tol=0.01)

    result = humidex(tdb=43, rh=20, model="masterson")
    assert math.isclose(result.humidex, 47.1, abs_tol=0.01)

    result = humidex(tdb=30, rh=30, model="masterson")
    assert math.isclose(result.humidex, 31.5, abs_tol=0.01)

    result = humidex(tdb=31, rh=55, model="masterson")
    assert math.isclose(result.humidex, 39.3, abs_tol=0.01)



import numpy as np

from pythermalcomfort.models import solar_gain
from tests.conftest import Urls, retrieve_reference_table, validate_result


def test_solar_gain(get_test_url, retrieve_data) -> None:
    """Test that the solar gain function calculates correctly for various inputs."""
    reference_table = retrieve_reference_table(
        get_test_url,
        retrieve_data,
        Urls.SOLAR_GAIN.name,
    )
    tolerance = reference_table["tolerance"]

    for entry in reference_table["data"]:
        inputs = entry["inputs"]
        outputs = entry["outputs"]
        result = solar_gain(**inputs)

        validate_result(result, outputs, tolerance)


def test_solar_gain_array() -> None:
    """Test that the solar gain function works with arrays."""
    np.allclose(
        solar_gain(
            sol_altitude=[0, 30],
            sharp=[120, 60],
            sol_radiation_dir=[800, 600],
            sol_transmittance=[0.5, 0.6],
            f_svv=[0.5, 0.4],
            f_bes=[0.5, 0.6],
            asw=0.7,
            posture="sitting",
        ).erf,
        np.array([46.4, 52.8]),
        atol=0.1,
    )



import numpy as np
from numba import float64, vectorize

from pythermalcomfort.utilities import met_to_w_m2


@vectorize(
    [
        float64(
            float64,
            float64,
            float64,
            float64,
            float64,
            float64,
            float64,
        ),
    ],
    cache=True,
)
def _pmv_ppd_optimized(tdb, tr, vr, rh, met, clo, wme):
    pa = rh * 10 * np.exp(16.6536 - 4030.183 / (tdb + 235))

    icl = 0.155 * clo  # thermal insulation of the clothing in M2K/W
    m = met * met_to_w_m2  # metabolic rate in W/M2
    w = wme * met_to_w_m2  # external work in W/M2
    mw = m - w  # internal heat production in the human body
    # calculation of the clothing area factor
    f_cl = (
        1 + 1.29 * icl if icl <= 0.078 else 1.05 + 0.645 * icl
    )  # ratio of surface clothed body over nude body

    # heat transfer coefficient by forced convection
    hcf = 12.1 * np.sqrt(vr)
    hc = hcf  # initialize variable
    taa = tdb + 273
    tra = tr + 273
    t_cla = taa + (35.5 - tdb) / (3.5 * icl + 0.1)

    p1 = icl * f_cl
    p2 = p1 * 3.96
    p3 = p1 * 100
    p4 = p1 * taa
    p5 = (308.7 - 0.028 * mw) + (p2 * (tra / 100.0) ** 4)
    xn = t_cla / 100
    xf = t_cla / 50
    eps = 0.00015

    n = 0
    while np.abs(xn - xf) > eps:
        xf = (xf + xn) / 2
        hcn = 2.38 * np.abs(100.0 * xf - taa) ** 0.25
        hc = max(hcn, hcf)
        xn = (p5 + p4 * hc - p2 * xf**4) / (100 + p3 * hc)
        n += 1
        if n > 150:
            raise StopIteration("Max iterations exceeded")

    tcl = 100 * xn - 273

    # heat loss diff. through skin
    hl1 = 3.05 * 0.001 * (5733 - (6.99 * mw) - pa)
    # heat loss by sweating
    hl2 = 0.42 * (mw - met_to_w_m2) if mw > met_to_w_m2 else 0
    # latent respiration heat loss
    hl3 = 1.7 * 0.00001 * m * (5867 - pa)
    # dry respiration heat loss
    hl4 = 0.0014 * m * (34 - tdb)
    # heat loss by radiation
    hl5 = 3.96 * f_cl * (xn**4 - (tra / 100.0) ** 4)
    # heat loss by convection
    hl6 = f_cl * hc * (tcl - tdb)

    ts = 0.303 * np.exp(-0.036 * m) + 0.028
    _pmv = ts * (mw - hl1 - hl2 - hl3 - hl4 - hl5 - hl6)

    return _pmv



from __future__ import annotations

import numpy as np
import numpy.typing as npt

from pythermalcomfort.classes_input import WBGTInputs
from pythermalcomfort.classes_return import WBGT


def wbgt(
    twb: float | npt.ArrayLike,
    tg: float | npt.ArrayLike,
    tdb: float | npt.ArrayLike = None,
    with_solar_load: bool = False,
    round_output: bool = True,
) -> WBGT:
    """Calculate the Wet Bulb Globe Temperature (WBGT) index in compliance
    with the ISO 7243 Standard [7243ISO2017]_.

    The WBGT is a heat stress index that
    measures the thermal environment to which a person is exposed. In most
    situations, this index is simple to calculate. It should be used as a
    screening tool to determine whether heat stress is present. The PHS model
    allows a more accurate estimation of stress. PHS can be calculated using
    the function :py:meth:`pythermalcomfort.models.phs`.

    The WBGT determines the impact of heat on a person throughout the course of a working
    day (up to 8 hours). It does not apply to very brief heat exposures. It pertains to
    the evaluation of male and female people who are fit for work in both indoor and
    outdoor occupational environments, as well as other sorts of surroundings [7243ISO2017]_.

    The WBGT is defined as a function of only twb and tg if the person is not exposed to
    direct radiant heat from the sun. When a person is exposed to direct radiant heat,
    tdb must also be specified.

    Parameters
    ----------
    twb : float or list of floats
        Natural (no forced air flow) wet bulb temperature, [°C].
    tg : float or list of floats
        Globe temperature, [°C].
    tdb : float or list of floats, optional
        Dry bulb air temperature, [°C]. This value is needed as input if the person is
        exposed to direct solar radiation.
    with_solar_load : bool, optional
        True if the globe sensor is exposed to direct solar radiation. Defaults to False.
    round_output : bool, optional
        If True, rounds output value. If False, it does not round it. Defaults to True.

    Returns
    -------
    WBGT
        A dataclass containing the Wet Bulb Globe Temperature Index. See
        :py:class:`~pythermalcomfort.classes_return.WBGT` for more details. To access the
        `wbgt` value, use the `wbgt` attribute of the returned `wbgt` instance, e.g.,
        `result.wbgt`.

    Examples
    --------
    .. code-block:: python

        from pythermalcomfort.models import wbgt

        result = wbgt(twb=25, tg=32)
        print(result.wbgt)  # 27.1

        result = wbgt(twb=25, tg=32, tdb=20, with_solar_load=True)
        print(result.wbgt)  # 25.9

    """
    # Validate inputs using the WBGTInputs class
    WBGTInputs(
        twb=twb,
        tg=tg,
        tdb=tdb,
        with_solar_load=with_solar_load,
        round_output=round_output,
    )

    twb = np.array(twb)
    tg = np.array(tg)
    tdb = np.array(tdb) if tdb is not None else None

    if with_solar_load and tdb is None:
        raise ValueError("Please enter the dry bulb air temperature")

    if with_solar_load:
        t_wbg = 0.7 * twb + 0.2 * tg + 0.1 * tdb
    else:
        t_wbg = 0.7 * twb + 0.3 * tg

    if round_output:
        t_wbg = np.round(t_wbg, 1)

    return WBGT(wbgt=t_wbg)



from __future__ import annotations

import numpy as np

from pythermalcomfort.classes_input import VerticalTGradPPDInputs
from pythermalcomfort.classes_return import VerticalTGradPPD
from pythermalcomfort.models import pmv_ppd_ashrae
from pythermalcomfort.utilities import Models, _check_standard_compliance_array


def vertical_tmp_grad_ppd(
    tdb: float | list[float],
    tr: float | list[float],
    vr: float | list[float],
    rh: float | list[float],
    met: float | list[float],
    clo: float | list[float],
    vertical_tmp_grad: float | list[float],
    round_output: bool = True,
) -> VerticalTGradPPD:
    """Calculate the percentage of thermally dissatisfied people with a
    vertical temperature gradient between feet and head [55ASHRAE2023]_. This equation is
    only applicable for vr < 0.2 m/s (40 fps).

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, [°C].

        .. note::
            The air temperature is the average value over two heights: 0.6 m (24 in.)
            and 1.1 m (43 in.) for seated occupants
            and 1.1 m (43 in.) and 1.7 m (67 in.) for standing occupants.

    tr : float or list of floats
        Mean radiant temperature, [°C].
    vr : float or list of floats
        Relative air speed, [m/s].

        .. note::
            vr is the relative air speed caused by body movement and not the air
            speed measured by the air speed sensor. The relative air speed is the sum of the
            average air speed measured by the sensor plus the activity-generated air speed
            (Vag). Where Vag is the activity-generated air speed caused by motion of
            individual body parts. vr can be calculated using the function
            :py:meth:`pythermalcomfort.utilities.v_relative`.

    rh : float or list of floats
        Relative humidity, [%].
    met : float or list of floats
        Metabolic rate, [met].
    clo : float or list of floats
        Clothing insulation, [clo].

        .. note::
            this is the basic insulation also known as the intrinsic clothing insulation value of the
            clothing ensemble (`I`:sub:`cl,r`), this is the thermal insulation from the skin
            surface to the outer clothing surface, including enclosed air layers, under actual
            environmental conditions. This value is not the total insulation (`I`:sub:`T,r`).
            The dynamic clothing insulation, clo, can be calculated using the function
            :py:meth:`pythermalcomfort.utilities.clo_dynamic_ashrae`.

    vertical_tmp_grad : float or list of floats
        Vertical temperature gradient between the feet and the head, [°C/m].
    round_output : bool, optional
        If True, rounds output value. If False, it does not round it. Defaults to True.

    Returns
    -------
    VerticalTGradPPD
        A dataclass containing the Predicted Percentage of Dissatisfied occupants with vertical temperature gradient and acceptability.
        See :py:class:`~pythermalcomfort.classes_return.VerticalTmpGradPPD` for more details.
        To access the `ppd_vg` and `acceptability` values, use the corresponding attributes of the returned `VerticalTmpGradPPD` instance, e.g., `result.ppd_vg`.

    Examples
    --------
    .. code-block:: python

        from pythermalcomfort.models import vertical_tmp_grad_ppd

        result = vertical_tmp_grad_ppd(
            tdb=25, tr=25, vr=0.1, rh=50, met=1.2, clo=0.5, vertical_tmp_grad=7
        )
        print(result.ppd_vg)  # 12.6
        print(result.acceptability)  # False

    """
    # Validate inputs using the VerticalTmpGradPPDInputs class
    VerticalTGradPPDInputs(
        tdb=tdb,
        tr=tr,
        vr=vr,
        rh=rh,
        met=met,
        clo=clo,
        vertical_tmp_grad=vertical_tmp_grad,
    )

    tdb = np.array(tdb)
    tr = np.array(tr)
    vr = np.array(vr)
    met = np.array(met)
    clo = np.array(clo)
    vertical_tmp_grad = np.array(vertical_tmp_grad)

    (
        tdb_valid,
        tr_valid,
        v_valid,
        met_valid,
        clo_valid,
    ) = _check_standard_compliance_array(
        standard=Models.ashrae_55_2023.value,
        tdb=tdb,
        tr=tr,
        v_limited=vr,
        met=met,
        clo=clo,
    )

    tsv = pmv_ppd_ashrae(
        tdb=tdb,
        tr=tr,
        vr=vr,
        rh=rh,
        met=met,
        clo=clo,
        model=Models.ashrae_55_2023.value,
    ).pmv
    numerator = np.exp(0.13 * (tsv - 1.91) ** 2 + 0.15 * vertical_tmp_grad - 1.6)
    ppd_val = (numerator / (1 + numerator) - 0.345) * 100
    acceptability = ppd_val <= 5

    if round_output:
        ppd_val = np.round(ppd_val, 1)

    all_valid = ~(
        np.isnan(tdb_valid)
        | np.isnan(tr_valid)
        | np.isnan(v_valid)
        | np.isnan(met_valid)
        | np.isnan(clo_valid)
    )

    ppd_val = np.where(all_valid, ppd_val, np.nan)
    acceptability = np.where(all_valid, acceptability, np.nan)

    return VerticalTGradPPD(ppd_vg=ppd_val, acceptability=acceptability)



import pytest

from pythermalcomfort.models import phs
from pythermalcomfort.utilities import met_to_w_m2
from tests.conftest import Urls, retrieve_reference_table, validate_result


def test_phs(get_test_url, retrieve_data) -> None:
    """Test that the function calculates the Predicted Heat Strain (PHS) correctly for various inputs."""
    reference_table = retrieve_reference_table(
        get_test_url,
        retrieve_data,
        Urls.PHS.name,
    )
    tolerance = reference_table["tolerance"]

    for entry in reference_table["data"]:
        inputs = entry["inputs"]
        inputs["model"] = "7933-2004"
        outputs = entry["outputs"]
        result = phs(**inputs)

        validate_result(result, outputs, tolerance)


weight = 75
height = 1.8
a_dubois = 0.202 * (weight**0.425) * (height**0.725)


@pytest.mark.parametrize(
    ("inputs", "expected"),
    [
        (
            {
                "tdb": 40,
                "tr": 40,
                "rh": 35,
                "v": 0.3,
                "met": 300 / met_to_w_m2 / a_dubois,
                "clo": 0.5,
                "posture": "standing",
                "wme": 0,
                "model": "7933-2023",
                "duration": 480,
                "limit_inputs": False,
                "acclimatized": 100,
                "round_output": False,
            },
            {
                "water_loss": 6538,
                "t_cr": 37.6,
                "d_lim_loss_95": 280,
            },
        ),
        (
            {
                "tdb": 35,
                "tr": 35,
                "rh": 60,
                "v": 0.1,
                "met": 300 / met_to_w_m2 / a_dubois,
                "clo": 0.5,
                "posture": "standing",
                "wme": 0,
                "model": "7933-2023",
                "duration": 480,
                "limit_inputs": False,
                "acclimatized": 0,
                "round_output": False,
            },
            {
                "water_loss": 6345,
                "t_cr": 40.8,
                "d_lim_loss_95": 250,
                "d_lim_t_re": 62,
            },
        ),
        pytest.param(
            {
                "tdb": 30,
                "tr": 54.2,
                "rh": 35,
                "v": 0.1,
                "met": 300 / met_to_w_m2 / a_dubois,
                "clo": 0.8,
                "posture": "standing",
                "wme": 0,
                "a_p": 0.3,
                "f_r": 0.85,
                "model": "7933-2023",
                "duration": 480,
                "limit_inputs": False,
                "acclimatized": 0,
                "round_output": False,
            },
            {
                "water_loss": 6419,
                "t_cr": 38.7,
                "d_lim_loss_95": 280,
                "d_lim_t_re": 149,
            },
            id="high-radiant-temp",
            marks=pytest.mark.xfail(
                reason="Known discrepancy t_cr and d_lim_t_re",
                strict=True,
            ),
        ),
        (
            {
                "tdb": 30,
                "tr": 30,
                "rh": 45,
                "v": 1.0,
                "met": 450 / met_to_w_m2 / a_dubois,
                "clo": 0.5,
                "posture": "standing",
                "wme": 0,
                "model": "7933-2023",
                "duration": 480,
                "limit_inputs": False,
                "acclimatized": 0,
                "round_output": False,
            },
            {
                "water_loss": 4593,
                "t_cr": 38.0,
                "d_lim_loss_95": 400,
            },
        ),
        (
            {
                "tdb": 35,
                "tr": 74.6,
                "rh": 30,
                "v": 1.0,
                "met": 250 / met_to_w_m2 / a_dubois,
                "clo": 1,
                "posture": "sitting",
                "wme": 0,
                "a_p": 0.2,
                "f_r": 0.85,
                "model": "7933-2023",
                "duration": 480,
                "limit_inputs": False,
                "acclimatized": 100,
                "round_output": False,
            },
            {
                "water_loss": 5813,
                "t_cr": 37.5,
                "d_lim_loss_95": 310,
            },
        ),
        pytest.param(
            {
                "tdb": [35, 35],
                "tr": 74.6,
                "rh": 30,
                "v": 1.0,
                "met": 250 / met_to_w_m2 / a_dubois,
                "clo": 1,
                "posture": "sitting",
                "wme": 0,
                "a_p": 0.2,
                "f_r": 0.85,
                "model": "7933-2023",
                "duration": 480,
                "limit_inputs": False,
                "acclimatized": 100,
                "round_output": False,
            },
            {
                "water_loss": [5813, 5813],
                "t_cr": [37.5, 37.5],
                "d_lim_loss_95": [310, 310],
            },
            id="array-input",
        ),
    ],
)
def test_2023_standard(inputs, expected) -> None:
    """Test the 2023 PHS model with various inputs."""
    result = phs(**inputs)
    assert result.water_loss == pytest.approx(expected["water_loss"], rel=0.025)
    assert result.t_cr == pytest.approx(expected["t_cr"], abs=0.3)
    assert result.d_lim_loss_95 == pytest.approx(expected["d_lim_loss_95"], abs=10)
    if "d_lim_t_re" in expected:
        assert result.d_lim_t_re == pytest.approx(expected["d_lim_t_re"], abs=10)


def test_value_acclimatized() -> None:
    """Test that the function raises ValueError for invalid acclimatized values."""
    with pytest.raises(ValueError):
        phs(
            tdb=40,
            tr=40,
            rh=33.85,
            v=0.3,
            met=2.58,
            clo=0.5,
            posture="standing",
            acclimatized=101,
        )

    with pytest.raises(ValueError):
        phs(
            tdb=40,
            tr=40,
            rh=33.85,
            v=0.3,
            met=2.58,
            clo=0.5,
            posture="standing",
            acclimatized=-1,
        )


def test_value_weight() -> None:
    """Test that the function raises a ValueError for invalid weight values."""
    with pytest.raises(ValueError):
        phs(
            tdb=40,
            tr=40,
            rh=33.85,
            v=0.3,
            met=2.58,
            clo=0.5,
            posture="standing",
            weight=1001,
        )

    with pytest.raises(ValueError):
        phs(
            tdb=40,
            tr=40,
            rh=33.85,
            v=0.3,
            met=2.58,
            clo=0.5,
            posture="standing",
            weight=0,
        )


def test_value_drink() -> None:
    """Test that drink input is within valid range."""
    with pytest.raises(ValueError):
        phs(
            tdb=40,
            tr=40,
            rh=33.85,
            v=0.3,
            met=2.58,
            clo=0.5,
            posture="standing",
            drink=0.5,
        )
    with pytest.raises(ValueError):
        phs(
            tdb=40,
            tr=40,
            rh=33.85,
            v=0.3,
            met=2.58,
            clo=0.5,
            posture="standing",
            drink=2,
        )



from __future__ import annotations

import numpy as np
import pytest

from pythermalcomfort.classes_input import WorkIntensity
from pythermalcomfort.classes_return import WorkCapacity
from pythermalcomfort.models import work_capacity_hothaps


def _expected_capacity(
    wbgt: float | list[float],
    divisor: float,
    exponent: float,
) -> np.ndarray:
    """Compute the expected work capacity based on WBGT, divisor, and exponent.

    Parameters
    ----------
    wbgt: float
        Wet Bulb Globe Temperature value(s).
    divisor:
        Divisor parameter for the capacity formula.
    exponent:
        Exponent parameter for the capacity formula.

    Returns
    -------
        Numpy array of capacity values, clipped between 0 and 100.

    """
    wbgt_array = np.asarray(wbgt)
    cap = 100 * (0.1 + 0.9 / (1 + (wbgt_array / divisor) ** exponent))
    return cap.clip(0, 100)


def test_scalar_heavy() -> None:
    """Test that a scalar heavy work intensity produces the expected capacity."""
    result = work_capacity_hothaps(30.0, work_intensity=WorkIntensity.HEAVY.value)
    assert isinstance(result, WorkCapacity)
    assert isinstance(result.capacity, float)
    expected = _expected_capacity(30.0, divisor=30.94, exponent=16.64)
    assert pytest.approx(expected, rel=1e-3) == result.capacity


def test_scalar_moderate() -> None:
    """Test that a scalar moderate work intensity produces the expected capacity."""
    result = work_capacity_hothaps(30.0, work_intensity=WorkIntensity.MODERATE.value)
    assert isinstance(result, WorkCapacity)
    assert isinstance(result.capacity, float)
    expected = _expected_capacity(30.0, divisor=32.93, exponent=17.81)
    assert pytest.approx(expected, rel=1e-3) == result.capacity


def test_scalar_light() -> None:
    """Test that a scalar light work intensity produces the expected capacity."""
    result = work_capacity_hothaps(30.0, work_intensity=WorkIntensity.LIGHT.value)
    assert isinstance(result, WorkCapacity)
    assert isinstance(result.capacity, float)
    expected = _expected_capacity(30.0, divisor=34.64, exponent=22.72)
    assert pytest.approx(expected, rel=1e-3) == result.capacity


def test_list_input() -> None:
    """Test that the function handles list inputs correctly."""
    wbgts = [20.0, 40.0]
    result = work_capacity_hothaps(wbgts, work_intensity="heavy")
    assert isinstance(result.capacity, np.ndarray)
    exp_list = _expected_capacity(wbgts, divisor=30.94, exponent=16.64).tolist()
    assert all(
        pytest.approx(e, rel=1e-3) == r
        for e, r in zip(exp_list, result.capacity, strict=False)
    )


def test_low_wbgt_clamped_to_100() -> None:
    """Test that the function clamps low WBGT values to 100% capacity."""
    result = work_capacity_hothaps(0.0, work_intensity="light")
    assert result.capacity == pytest.approx(100.0, rel=1e-6)


def test_high_wbgt_approaches_10_percent() -> None:
    """Test that the function returns approximately 10% capacity for high WBGT."""
    result = work_capacity_hothaps(100.0, work_intensity="moderate")
    assert result.capacity == pytest.approx(10.0, rel=1e-3)


def test_invalid_intensity_raises() -> None:
    """Test that the function raises ValueError for invalid work intensity."""
    with pytest.raises(ValueError):
        work_capacity_hothaps(30.0, work_intensity="invalid")



import numpy as np
import pytest

from pythermalcomfort.models import pet_steady
from tests.conftest import Urls, retrieve_reference_table, validate_result


def test_pet_steady(get_test_url, retrieve_data) -> None:
    """Test that the PET function calculates correctly for various inputs."""
    reference_table = retrieve_reference_table(
        get_test_url,
        retrieve_data,
        Urls.PET_STEADY.name,
    )
    tolerance = reference_table["tolerance"]

    for entry in reference_table["data"]:
        inputs = entry["inputs"]
        outputs = entry["outputs"]
        result = pet_steady(**inputs)

        validate_result(result, outputs, tolerance)


PET_TEST_MATRIX = (
    # 'tdb', 'tr', 'rh', 'v', 'met', 'clo', 'exp'
    (20, 20, 50, 0.15, 1.37, 0.5, 18.85),
    (30, 30, 50, 0.15, 1.37, 0.5, 30.6),
    (20, 20, 50, 0.5, 1.37, 0.5, 17.16),
    (21, 21, 50, 0.1, 1.37, 0.9, 21.08),
    (20, 20, 50, 0.1, 1.37, 0.9, 19.92),
    (-5, 40, 2, 0.5, 1.37, 0.9, 7.82),
    (-5, -5, 50, 5.0, 1.37, 0.9, -13.38),
    (30, 60, 80, 1.0, 1.37, 0.9, 44.63),
    (30, 30, 80, 1.0, 1.37, 0.9, 32.21),
)


@pytest.mark.parametrize("shape", [(10, 10), 10, (3, 3, 3)])
@pytest.mark.parametrize(("tdb", "tr", "rh", "v", "met", "clo", "exp"), PET_TEST_MATRIX)
def test_pet_array(shape, tdb, tr, rh, v, met, clo, exp) -> None:
    """Test that the PET function calculates correctly for various inputs."""
    tdb_arr = list(np.full(shape, tdb))
    tr_arr = list(np.full(shape, tr))
    rh_arr = list(np.full(shape, rh))
    v_arr = list(np.full(shape, v))
    res = pet_steady(tdb=tdb_arr, tr=tr_arr, rh=rh_arr, v=v_arr, met=met, clo=clo).pet
    np.testing.assert_array_equal(actual=res, desired=np.full(shape, exp))



import pytest

from pythermalcomfort.models import wbgt
from tests.conftest import Urls, retrieve_reference_table, validate_result


def test_wbgt_with_url_cases(get_test_url, retrieve_data) -> None:
    """Test that the function calculates the Wet Bulb Globe Temperature (WBGT) correctly for various inputs."""
    reference_table = retrieve_reference_table(
        get_test_url,
        retrieve_data,
        Urls.WBGT.name,
    )
    tolerance = reference_table["tolerance"]

    for entry in reference_table["data"]:
        inputs = entry["inputs"]
        outputs = entry["outputs"]
        result = wbgt(**inputs)

        validate_result(result, outputs, tolerance)


# Test wbgt value error
def test_wbgt() -> None:
    """Test that the function raises ValueError if twb is not provided."""
    with pytest.raises(ValueError):
        wbgt(twb=25, tg=32, with_solar_load=True)


#  Calculate WBGT with twb and tg set to None
def test_calculate_wbgt_with_twb_and_tg_set_to_none() -> None:
    """Test that the function raises Error with twb and tg set to None."""
    with pytest.raises(TypeError):
        wbgt(None, None)



from __future__ import annotations

import math
import warnings
from enum import Enum
from typing import NamedTuple

import numpy as np

from pythermalcomfort.classes_return import PsychrometricValues
from pythermalcomfort.shared_functions import valid_range

warnings.simplefilter("always")


c_to_k = 273.15
cp_vapour = 1805.0
cp_water = 4186
cp_air = 1004
h_fg = 2501000
r_air = 287.055
g = 9.81  # m/s2
met_to_w_m2 = 58.15


class Models(Enum):
    """Models options."""

    ashrae_55_2023 = "55-2023"
    iso_7730_2005 = "7730-2005"
    iso_9920_2007 = "9920-2007"
    iso_7933_2004 = "7933-2004"
    iso_7933_2023 = "7933-2023"


class Units(Enum):
    """Units options."""

    SI = "SI"
    IP = "IP"


class Sex(Enum):
    """Sex options."""

    male = "male"
    female = "female"


def p_sat_torr(tdb: float | list[float]) -> float | list[float]:
    """Estimates the saturation vapor pressure in [torr].

    Parameters
    ----------
    tdb : float or list of floats
        dry bulb air temperature, [C]

    Returns
    -------
    p_sat : float
        saturation vapor pressure [torr]

    """
    return np.exp(18.6686 - 4030.183 / (tdb + 235.0))


def enthalpy_air(
    tdb: float | list[float],
    hr: float | list[float],
) -> float | list[float]:
    """Calculate air enthalpy_air.

    Parameters
    ----------
    tdb: float or list of floats
        air temperature, [°C]
    hr: float or list of floats
        humidity ratio, [kg water/kg dry air]

    Returns
    -------
    enthalpy_air: float or list of floats
        enthalpy_air [J/kg dry air]

    """
    h_dry_air = cp_air * tdb
    h_sat_vap = h_fg + cp_vapour * tdb
    return h_dry_air + hr * h_sat_vap


# pre-calculated constants for p_sat
c1 = -5674.5359
c2 = 6.3925247
c3 = -0.9677843 * 1e-2
c4 = 0.62215701 * 1e-6
c5 = 0.20747825 * 1e-8
c6 = -0.9484024 * 1e-12
c7 = 4.1635019
c8 = -5800.2206
c9 = 1.3914993
c10 = -0.048640239
c11 = 0.41764768 * 1e-4
c12 = -0.14452093 * 1e-7
c13 = 6.5459673


def p_sat(tdb: float | list[float]) -> float | list[float]:
    """Calculate vapour pressure of water at different temperatures.

    Parameters
    ----------
    tdb: float or list of floats
        air temperature, [°C]

    Returns
    -------
    p_sat: float or list of floats
        saturation vapor pressure, [Pa]

    """
    ta_k = tdb + c_to_k
    # pre-calculate the value before passing it to .where
    log_ta_k = np.log(ta_k)
    return np.where(
        ta_k < c_to_k,
        np.exp(
            c1 / ta_k
            + c2
            + ta_k * (c3 + ta_k * (c4 + ta_k * (c5 + c6 * ta_k)))
            + c7 * log_ta_k,
        ),
        np.exp(
            c8 / ta_k + c9 + ta_k * (c10 + ta_k * (c11 + ta_k * c12)) + c13 * log_ta_k,
        ),
    )


def antoine(tdb: float | np.ndarray) -> np.ndarray:
    """Calculate saturated vapor pressure using Antoine equation [kPa].

    Parameters
    ----------
    tdb : float or list of floats
        Temperature [°C].

    Returns
    -------
    float or list of floats
        Saturated vapor pressure [kPa].

    """
    tdb = np.array(tdb)
    return math.e ** (16.6536 - 4030.183 / (tdb + 235))


def psy_ta_rh(
    tdb: float | list[float],
    rh: float | list[float],
    p_atm=101325,
) -> PsychrometricValues:
    """Calculate psychrometric values of air based on dry bulb air temperature and
    relative humidity.

    For more accurate results we recommend the use of the Python
    package `psychrolib`_.

    .. _psychrolib: https://pypi.org/project/PsychroLib/

    Parameters
    ----------
    tdb: float or list of floats
        air temperature, [°C]
    rh: float or list of floats
        relative humidity, [%]
    p_atm: float or list of floats
        atmospheric pressure, [Pa]

    Returns
    -------
    p_vap: float or list of floats
        partial pressure of water vapor in moist air, [Pa]
    hr: float or list of floats
        humidity ratio, [kg water/kg dry air]
    wet_bulb_tmp: float or list of floats
        wet bulb temperature, [°C]
    dew_point_tmp: float or list of floats
        dew point temperature, [°C]
    h: float or list of floats
        enthalpy_air [J/kg dry air]

    """
    tdb = np.array(tdb)
    rh = np.array(rh)

    p_saturation = p_sat(tdb)
    p_vap = rh / 100 * p_saturation
    hr = 0.62198 * p_vap / (p_atm - p_vap)
    tdp = dew_point_tmp(tdb, rh)
    twb = wet_bulb_tmp(tdb, rh)
    h = enthalpy_air(tdb, hr)

    return PsychrometricValues(
        p_sat=p_saturation,
        p_vap=p_vap,
        hr=hr,
        wet_bulb_tmp=twb,
        dew_point_tmp=tdp,
        h=h,
    )


def wet_bulb_tmp(
    tdb: float | list[float],
    rh: float | list[float],
) -> float | list[float]:
    """Calculate the wet-bulb temperature using the Stull equation [Stull2011]_.

    Parameters
    ----------
    tdb: float or list of floats
        air temperature, [°C]
    rh: float or list of floats
        relative humidity, [%]

    Returns
    -------
    tdb: float or list of floats
        wet-bulb temperature, [°C]

    """
    twb = (
        tdb * np.arctan(0.151977 * (rh + 8.313659) ** 0.5)
        + np.arctan(tdb + rh)
        - np.arctan(rh - 1.676331)
        + 0.00391838 * rh**1.5 * np.arctan(0.023101 * rh)
        - 4.686035
    )

    return np.around(twb, 1)


def dew_point_tmp(
    tdb: float | list[float],
    rh: float | list[float],
) -> float | list[float]:
    """Calculate the dew point temperature.

    Parameters
    ----------
    tdb: float or list of floats
        dry bulb air temperature, [°C]
    rh: float or list of floats
        relative humidity, [%]

    Returns
    -------
    dew_point_tmp: float or list of floats
        dew point temperature, [°C]

    """
    tdb = np.array(tdb)
    rh = np.array(rh)

    c = 257.14
    b = 18.678
    d = 234.5

    gamma_m = np.log(rh / 100 * np.exp((b - tdb / d) * (tdb / (c + tdb))))

    return np.round(c * gamma_m / (b - gamma_m), 1)


def mean_radiant_tmp(
    tg: float | list[float],
    tdb: float | list[float],
    v: float | list[float],
    d: float | list[float] = 0.15,
    emissivity: float | list[float] = 0.95,
    standard="Mixed Convection",
) -> float | list[float]:
    """Convert the globe temperature reading into mean radiant temperature in accordance
    with either the Mixed Convection developed by Teitelbaum E. et al. (2022) or the ISO
    7726:1998 Standard [7726ISO1998]_.

    Parameters
    ----------
    tg : float or list of floats
        globe temperature, [°C]
    tdb : float or list of floats
        air temperature, [°C]
    v : float or list of floats
        air speed, [m/s]
    d : float or list of floats
        diameter of the globe, [m] default 0.15 m
    emissivity : float or list of floats
        emissivity of the globe temperature sensor, default 0.95
    standard : str, optional
        Supported values are 'Mixed Convection' and 'ISO'. Defaults to 'Mixed Convection'.
        either choose between the Mixed Convection and ISO formulations.
        The Mixed Convection formulation has been proposed by Teitelbaum E. et al. (2022)
        to better determine the free and forced convection coefficient used in the
        calculation of the mean radiant temperature. They also showed that mean radiant
        temperature measured with ping-pong ball-sized globe thermometers is not reliable
        due to a stochastic convective bias [Teitelbaum2022]_. The Mixed Convection model has only
        been validated for globe sensors with a diameter between 0.04 and 0.15 m.

    Returns
    -------
    tr: float or list of floats
        mean radiant temperature, [°C]

    """
    standard = standard.lower()

    tdb = np.array(tdb)
    tg = np.array(tg)
    v = np.array(v)
    d = np.array(d)

    if standard == "mixed convection":
        mu = 0.0000181  # Pa s
        k_air = 0.02662  # W/m-K
        beta = 0.0034  # 1/K
        nu = 0.0000148  # m2/s
        alpha = 0.00002591  # m2/s
        pr = cp_air * mu / k_air  # Prandtl constants

        o = 0.0000000567
        n = 1.27 * d + 0.57

        ra = g * beta * np.absolute(tg - tdb) * d * d * d / nu / alpha
        re = v * d / nu

        nu_natural = 2 + (0.589 * np.power(ra, (1 / 4))) / (
            np.power(1 + np.power(0.469 / pr, 9 / 16), (4 / 9))
        )
        nu_forced = 2 + (
            0.4 * np.power(re, 0.5) + 0.06 * np.power(re, 2 / 3)
        ) * np.power(pr, 0.4)

        tr = (
            np.power(
                np.power(tg + 273.15, 4)
                - (
                    (
                        (
                            (
                                np.power(
                                    (np.power(nu_forced, n) + np.power(nu_natural, n)),
                                    1 / n,
                                )
                            )
                            * k_air
                            / d
                        )
                        * (-tg + tdb)
                    )
                    / emissivity
                    / o
                ),
                0.25,
            )
            - 273.15
        )

        d_valid = valid_range(d, (0.04, 0.15))
        tr = np.where(~np.isnan(d_valid), tr, np.nan)

        return np.around(tr, 1)

    if standard == "iso":  # pragma: no branch
        tg = np.add(tg, c_to_k)
        tdb = np.add(tdb, c_to_k)

        # calculate heat transfer coefficient
        h_n = 1.4 * np.power(np.abs(tg - tdb) / d, 0.25)  # natural convection
        h_f = 6.3 * np.power(v, 0.6) / np.power(d, 0.4)  # forced convection

        # get the biggest between the two coefficients
        h = np.maximum(h_f, h_n)

        tr = (
            np.power(
                np.power(tg, 4) + h * (tg - tdb) / (emissivity * (5.67 * 10**-8)),
                0.25,
            )
            - c_to_k
        )

        return np.around(tr, 1)


def validate_type(value, name: str, allowed_types: tuple):
    """Validate the type of a value against allowed types."""
    if isinstance(value, np.generic):
        value = value.item()
    if not isinstance(value, allowed_types):
        invalid_type_msg = (
            f"{name} must be one of the following types: {allowed_types}."
        )
        raise TypeError(invalid_type_msg)


def transpose_sharp_altitude(sharp: float, altitude: float) -> float:
    """Transpose the solar altitude and solar azimuth angles."""
    altitude_new = math.degrees(
        math.asin(
            math.sin(math.radians(abs(sharp - 90))) * math.cos(math.radians(altitude)),
        ),
    )
    sharp = math.degrees(
        math.atan(
            math.sin(math.radians(sharp)) * math.tan(math.radians(90 - altitude)),
        ),
    )
    sol_altitude = altitude_new
    return round(sharp, 3), round(sol_altitude, 3)


def _check_standard_compliance_array(standard, **kwargs):
    default_kwargs = {"airspeed_control": True}
    params = {**default_kwargs, **kwargs}
    values_to_return = {}

    if standard == Models.ashrae_55_2023.value:  # based on table 7.3.4 ashrae 55 2020
        tdb_valid = valid_range(params["tdb"], (10.0, 40.0))
        tr_valid = valid_range(params["tr"], (10.0, 40.0))

        values_to_return["tdb"] = tdb_valid
        values_to_return["tr"] = tr_valid

        if "v" in params:
            v_valid = valid_range(params["v"], (0.0, 2.0))
            values_to_return["v"] = v_valid

        if not params["airspeed_control"]:
            v_valid = np.where(
                (params["v"] > 0.8) & (params["clo"] < 0.7) & (params["met"] < 1.3),
                np.nan,
                v_valid,
            )
            to = operative_tmp(params["tdb"], params["tr"], params["v"])
            v_limit = 50.49 - 4.4047 * to + 0.096425 * to * to
            v_valid = np.where(
                (to > 23)
                & (to < 25.5)
                & (params["v"] > v_limit)
                & (params["clo"] < 0.7)
                & (params["met"] < 1.3),
                np.nan,
                v_valid,
            )
            v_valid = np.where(
                (to <= 23)
                & (params["v"] > 0.2)
                & (params["clo"] < 0.7)
                & (params["met"] < 1.3),
                np.nan,
                v_valid,
            )

            values_to_return["v"] = v_valid

        if "met" in params:
            met_valid = valid_range(params["met"], (1.0, 4.0))
            clo_valid = valid_range(params["clo"], (0.0, 1.5))

            values_to_return["met"] = met_valid
            values_to_return["clo"] = clo_valid

        if "v_limited" in params:
            valid = valid_range(params["v_limited"], (0.0, 0.2))
            values_to_return["v_limited"] = valid

        return values_to_return.values()

    if standard == Models.iso_7933_2004.value:  # based on ISO 7933:2004 Annex A
        tdb_valid = valid_range(params["tdb"], (15.0, 50.0))
        p_a_valid = valid_range(params["p_a"], (0, 4.5))
        tr_valid = valid_range(params["tr"], (0.0, 60.0))
        v_valid = valid_range(params["v"], (0.0, 3))
        met_valid = valid_range(params["met"], (100, 450))
        clo_valid = valid_range(params["clo"], (0.1, 1))

        return tdb_valid, tr_valid, v_valid, p_a_valid, met_valid, clo_valid

    if standard == Models.iso_7933_2023.value:  # based on ISO 7933:2023 Annex A
        tdb_valid = valid_range(params["tdb"], (15.0, 50.0))
        p_a_valid = valid_range(params["p_a"], (0.5, 4.5))
        tr_valid = valid_range(params["tr"], (0.0, 60.0))
        v_valid = valid_range(params["v"], (0.0, 3))
        met_valid = valid_range(params["met"], (100, 450))
        clo_valid = valid_range(params["clo"], (0.1, 1))

        return tdb_valid, tr_valid, v_valid, p_a_valid, met_valid, clo_valid

    if standard == "fan_heatwaves":
        tdb_valid = valid_range(params["tdb"], (20.0, 50.0))
        tr_valid = valid_range(params["tr"], (20.0, 50.0))
        v_valid = valid_range(params["v"], (0.1, 4.5))
        rh_valid = valid_range(params["rh"], (0, 100))
        met_valid = valid_range(params["met"], (0.7, 2))
        clo_valid = valid_range(params["clo"], (0.0, 1))

        return tdb_valid, tr_valid, v_valid, rh_valid, met_valid, clo_valid

    if standard == Models.iso_7730_2005.value:  # based on ISO 7730:2005 page 3
        tdb_valid = valid_range(params["tdb"], (10.0, 30.0))
        tr_valid = valid_range(params["tr"], (10.0, 40.0))
        v_valid = valid_range(params["v"], (0.0, 1.0))
        met_valid = valid_range(params["met"], (0.8, 4.0))
        clo_valid = valid_range(params["clo"], (0.0, 2))

        return tdb_valid, tr_valid, v_valid, met_valid, clo_valid


class Postures(Enum):
    """Postures options."""

    standing = "standing"
    sitting = "sitting"
    sedentary = "sedentary"
    reclining = "reclining"
    lying = "lying"
    supine = "supine"
    crouching = "crouching"


class BodySurfaceAreaEquations(Enum):
    """Body Surface Area Equations."""

    dubois = "dubois"
    takahira = "takahira"
    fujimoto = "fujimoto"
    kurazumi = "kurazumi"


def body_surface_area(
    weight: float,
    height: float,
    formula: str = BodySurfaceAreaEquations.dubois.value,
) -> float:
    """Calculate the body surface area (BSA) in square meters.

    Parameters
    ----------
    weight : float
        Body weight in kilograms.
    height : float
        Body height in meters.
    formula : str, optional
        Formula used to calculate the body surface area. Default is "dubois".
        Choose one from BodySurfaceAreaEquations.

    Returns
    -------
    float
        Body surface area in square meters.

    Raises
    ------
    ValueError
        If the specified formula is not recognized.

    Examples
    --------
    Calculate BSA using the DuBois formula:

    .. code-block:: python

        bsa = body_surface_area(weight=70, height=1.75, formula="dubois")
        print(bsa)

    """
    if formula == BodySurfaceAreaEquations.dubois.value:
        return 0.202 * (weight**0.425) * (height**0.725)
    if formula == BodySurfaceAreaEquations.takahira.value:
        return 0.2042 * (weight**0.425) * (height**0.725)
    if formula == BodySurfaceAreaEquations.fujimoto.value:
        return 0.1882 * (weight**0.444) * (height**0.663)
    if formula == BodySurfaceAreaEquations.kurazumi.value:
        return 0.2440 * (weight**0.383) * (height**0.693)
    invalid_formula_msg = (
        f"Formula '{formula}' for calculating body surface area is not recognized."
    )
    raise ValueError(invalid_formula_msg)


def f_svv(w: float, h: float, d: float) -> float:
    """Calculate the sky-vault view fraction.

    Parameters
    ----------
    w : float
        width of the window, [m]
    h : float
        height of the window, [m]
    d : float
        distance between the occupant and the window, [m]

    Returns
    -------
    f_svv  : float
        sky-vault view fraction ranges between 0 and 1

    """
    return (
        math.degrees(math.atan(h / (2 * d)))
        * math.degrees(math.atan(w / (2 * d)))
        / 16200
    )


def v_relative(v: float | list[float], met: float | list[float]) -> np.ndarray:
    """Estimates the relative air speed which combines the average air speed of the
    space plus the relative air speed caused by the body movement. The same equation is
    used in the ASHRAE 55:2023 and ISO 7730:2005 standards.

    Parameters
    ----------
    v : float or list of floats
        air speed measured by the sensor, [m/s]
    met : float or list of floats
        metabolic rate, [met]

    Returns
    -------
    vr  : float or list of floats
        relative air speed, [m/s]

    """
    v = np.array(v)
    met = np.array(met)
    return np.where(met > 1, np.around(v + 0.3 * (met - 1), 3), v)


def clo_dynamic_ashrae(
    clo: float | list[float],
    met: float | list[float],
    model: str = Models.ashrae_55_2023.value,
) -> np.ndarray:
    """Estimates the dynamic intrinsic clothing insulation (I :sub:`cl,r`). The ASHRAE
    55:2023 refers to it as (I :sub:`cl,active`). The activity as well as the air speed
    modify the insulation characteristics of the clothing. Consequently, the ASHRAE 55
    standard provides a correction factor for the clothing insulation (I :sub:`cl`)
    based on the metabolic rate.

    Parameters
    ----------
    clo : float or list of floats
        clothing insulation, [clo]

        .. note::
            this is the basic insulation (I :sub:`cl`) also known as the intrinsic
            clothing insulation value under reference conditions

    met : float or list of floats
        metabolic rate, [met]
    model : str, optional
        Select the version of the ASHRAE 55 Standard to use. Currently, the only
        option available is "55-2023".

    Returns
    -------
    clo : float or list of floats
        dynamic clothing insulation (I :sub:`cl,r`), [clo]

    """
    clo = np.array(clo)
    met = np.array(met)

    model = model.lower()
    if model not in [Models.ashrae_55_2023.value]:
        invalid_model_msg = f"PMV calculations can only be performed in compliance with ASHRAE {Models.ashrae_55_2023.value}"
        raise ValueError(invalid_model_msg)

    return np.where(met > 1.2, np.around(clo * (0.6 + 0.4 / met), 3), clo)


def clo_dynamic_iso(
    clo: float | list[float],
    met: float | list[float],
    v: float | list[float],
    i_a: float | list[float] = 0.7,
    model: str = Models.iso_9920_2007.value,
) -> np.ndarray:
    """Estimates the dynamic intrinsic clothing insulation (I :sub:`cl,r`). The activity
    as well as the air speed modify the insulation characteristics of the clothing.
    Consequently, the ISO standard states that (I :sub:`cl,`) shall be corrected
    [7730ISO2005]_. However, the ISO 7730:2005 contains insufficient information to
    calculate (I :sub:`cl,r`). Therefore, we implemented the equations provided in the
    ISO 9920:2007 standard [ISO9920]_.

    Parameters
    ----------
    clo : float or list of floats
        clothing insulation, [clo]
    met : float or list of floats
        metabolic rate, [met]
    v : float or list of floats
        air speed, [m/s]
    i_a : float or list of floats
        thermal insulation of the boundary (surface) air layer around the outer clothing
        or, when nude, around the skin surface, [clo]
    model : str, optional
        Select the version of the ISO standard to use. Currently, the only
        option available is "9920-2007".

    Returns
    -------
    clo : float or list of floats
        dynamic clothing insulation, [clo]

    """
    model = model.lower()
    if model not in [Models.iso_9920_2007.value]:
        invalid_model_msg = f"PMV calculations can only be performed in compliance with ISO {Models.iso_9920_2007.value}"
        raise ValueError(invalid_model_msg)

    clo = np.array(clo)
    met = np.array(met)
    i_a = np.array(i_a)
    v = np.array(v)

    f_cl = clo_area_factor(i_cl=clo)
    i_t = clo + i_a / f_cl
    v_walk = v_relative(v=v, met=met) - v
    v_r = v_relative(v=v, met=met)
    i_t_r = clo_total_insulation(
        i_t=i_t,
        vr=v_r,
        v_walk=v_walk,
        i_a_static=i_a,
        i_cl=clo,
    )
    i_a_r = clo_insulation_air_layer(vr=v_r, v_walk=v_walk, i_a_static=i_a)
    return i_t_r - i_a_r / f_cl


def running_mean_outdoor_temperature(
    temp_array: list[float],
    alpha: float = 0.8,
    units: str = Units.SI.value,
) -> float:
    """Estimate the running mean temperature also known as prevailing mean outdoor
    temperature.

    Parameters
    ----------
    temp_array: list
        array containing the mean daily temperature in descending order (i.e. from
        newest/yesterday to oldest) :math:`[t_{day-1}, t_{day-2}, ... ,
        t_{day-n}]`.
        Where :math:`t_{day-1}` is yesterday's daily mean temperature. The EN
        16798-1 2019 [16798EN2019]_ states that n should be equal to 7
    alpha : float
        constant between 0 and 1. The EN 16798-1 2019 [16798EN2019]_ recommends a value of 0.8,
        while the ASHRAE 55 2020 recommends to choose values between 0.9 and 0.6,
        corresponding to a slow- and fast- response running mean, respectively.
        Adaptive comfort theory suggests that a slow-response running mean (alpha =
        0.9) could be more appropriate for climates in which synoptic-scale (day-to-
        day) temperature dynamics are relatively minor, such as the humid tropics.
    units: str default="SI"
        select the SI (International System of Units) or the IP (Imperial Units) system.

    Returns
    -------
    t_rm  : float
        running mean outdoor temperature

    """
    units = units.upper()
    if units == Units.IP.value:
        for ix, _x in enumerate(temp_array):
            temp_array[ix] = units_converter(tdb=temp_array[ix])[0]

    coeff = [alpha**ix for ix, x in enumerate(temp_array)]
    t_rm = sum([a * b for a, b in zip(coeff, temp_array, strict=False)]) / sum(coeff)

    if units == Units.IP.value:
        t_rm = units_converter(tmp=t_rm, from_units=Units.SI.value.lower())[0]

    return round(t_rm, 1)


def units_converter(from_units=Units.IP.value, **kwargs) -> list[float]:
    """Convert IP values to SI units.

    Parameters
    ----------
    from_units: str
        specify system to convert from
    **kwargs : [t, v]

    Returns
    -------
    converted values in SI units

    """
    results = []
    from_units = from_units.upper()
    if from_units == Units.IP.value:
        for key, value in kwargs.items():
            if "tmp" in key or key == "tr" or key == "tdb":
                results.append((value - 32) * 5 / 9)
            if key in ["v", "vr", "vel"]:
                results.append(value / 3.281)
            if key == "area":
                results.append(value / 10.764)
            if key == "pressure":
                results.append(value * 101325)

    elif from_units == Units.SI.value:
        for key, value in kwargs.items():
            if "tmp" in key or key == "tr" or key == "tdb":
                results.append((value * 9 / 5) + 32)
            if key in ["v", "vr", "vel"]:
                results.append(value * 3.281)
            if key == "area":
                results.append(value * 10.764)
            if key == "pressure":
                results.append(value / 101325)

    return results


def operative_tmp(
    tdb: float | list[float],
    tr: float | list[float],
    v: float | list[float],
    standard: str = "ISO",
) -> float | list[float]:
    """Calculate the operative temperature in accordance with ISO 7726:1998 [7726ISO1998]_.

    Parameters
    ----------
    tdb: float or list of floats
        air temperature, [°C]
    tr: float or list of floats
        mean radiant temperature, [°C]
    v: float or list of floats
        air speed, [m/s]
    standard: str (default="ISO")
        either choose between ISO and ASHRAE


    Returns
    -------
    to: float
        operative temperature, [°C]

    """
    if standard.lower() == "iso":
        return (tdb * np.sqrt(10 * v) + tr) / (1 + np.sqrt(10 * v))
    if standard.lower() == "ashrae":
        a = np.where(v < 0.6, 0.6, 0.7)
        a = np.where(v < 0.2, 0.5, a)
        return a * tdb + (1 - a) * tr
    error_message = (
        f"Operative temperature can only be calculated in compliance with ISO or ASHRAE standards. "
        f"Received standard: {standard}"
    )
    raise ValueError(error_message)


def clo_intrinsic_insulation_ensemble(
    clo_garments: float | list[float],
) -> float | list[float]:
    """Calculate the intrinsic insulation of a clothing ensemble based on individual
    garments.

    This equation is in accordance with the ISO 9920:2009 standard [ISO9920]_
    Section 4.3. It should be noted that this equation is only valid for clothing
    ensembles with rather uniform insulation values across the body.

    Parameters
    ----------
    clo_garments:  floats or list of floats
        list of floats containing the clothing insulation for each individual garment

    Returns
    -------
    i_cl: float
        intrinsic insulation of the clothing ensemble, [clo]

    """
    clo_garments = np.array(clo_garments)
    return np.sum(clo_garments) * 0.835 + 0.161


def clo_area_factor(i_cl: float | list[float]) -> float | list[float]:
    """Calculate the clothing area factor (f_cl) of the clothing ensemble as a function
    of the intrinsic insulation of the clothing ensemble. This equation is in accordance
    with the ISO 9920:2009 standard [ISO9920]_ Section 5. The standard warns that the
    correlation between f_cl and i_cl is low especially for non-western clothing
    ensembles. The application of this equation is limited to clothing ensembles with
    clo values between 0.2 and 1.7 clo.

    Parameters
    ----------
    i_cl: float or list of floats
        intrinsic insulation of the clothing ensemble, [clo]

    Returns
    -------
    f_cl: float or list of floats
        area factor of the clothing ensemble, [m2]

    """
    i_cl = np.array(i_cl)
    return 1 + 0.28 * i_cl


# TODO implement the vr and v_walk functions as a function of the met
def clo_insulation_air_layer(
    vr: float | list[float],
    v_walk: float | list[float],
    i_a_static: float | list[float],
) -> float | list[float]:
    """Calculate the insulation of the boundary air layer (`I`:sub:`a,r`).

    The static
    boundary air value is 0.7 clo (0.109 m2K/W) for air velocities around 0.1 m/s to
    0.15 m/s. Thus, for static conditions, the standard recommends using the value of
    0.7 clo (0.109 m2K/W) for the boundary air layer insulation. For walking conditions,
    the boundary air layer insulation is calculated based on the walking speed (v_walk)
    and the relative air speed (vr). This equation is extracted from the ISO 9920:2009
    standard [ISO9920]_ Section 6.

    Parameters
    ----------
    vr: float or list of floats
        relative air speed, [m/s]
    v_walk: float or list of floats
        walking speed, [m/s]
    i_a_static: float or list of floats
        static boundary air layer insulation, [clo]

    Returns
    -------
    i_a_r: float or list of floats
        boundary air layer insulation, [clo]

    """
    vr = np.array(vr)
    v_walk = np.array(v_walk)
    i_a_static = np.array(i_a_static)

    return (
        np.exp(
            -0.533 * (vr - 0.15)
            + 0.069 * (vr - 0.15) ** 2
            - 0.462 * v_walk
            + 0.201 * v_walk**2,
        )
        * i_a_static
    )


def clo_total_insulation(
    i_t: float | list[float],
    vr: float | list[float],
    v_walk: float | list[float],
    i_a_static: float | list[float],
    i_cl: float | list[float],
) -> float | list[float]:
    """Calculate the total insulation of the clothing ensemble (`I`:sub:`T,r`).

    The clothing ensemble (`I`:sub:`T,r`) which is
    the actual thermal insulation from the body surface to the environment, considering
    all clothing, enclosed air layers, and boundary air layers under given environmental
    conditions and activities. It accounts for the effects of movements and wind. The
    ISO 7790 standard [ISO9920]_ provides different equations to calculate it as a
    function of the total thermal insulation of clothing (`I`:sub:`T`), the insulation
    of the boundary air layer (`I`:sub:`a`), the walking speed (`v`:sub:`walk`), and the
    relative air speed (`v`:sub:`r`). These different equations are used if the person
    is clothed in normal clothing (0.6 clo < (`I`:sub:`cl`) < 1.4 clo or 1.2 clo <
    (`I`:sub:`T`) < 2.0 clo), nude (`I`:sub:`cl` = 0 clo), and if the person is clothed
    in very light clothing (`I`:sub:`cl` < 0.6 clo). Here we have not implemented the
    equation for high clothing (`I`:sub:`T` > 2.0 clo). Hence the applicability of this
    function is limited to 0 clo < (`I`:sub:`T`) < 2.0 clo). You can find all the inputs
    required in this function in the ISO 9920:2009 standard [ISO9920]_ Annex A.

    Parameters
    ----------
    i_t: float or list of floats
        total thermal insulation of clothing under static reference conditions [clo]
    vr: float or list of floats
        relative air speed, [m/s]
    v_walk: float or list of floats
        walking speed, [m/s]
    i_a_static: float or list of floats
        static boundary air layer insulation, [clo]
    i_cl: float or list of floats
        intrinsic insulation of the clothing ensemble, this is the thermal insulation
        from the skin surface to the outer clothing surface [clo]

    Returns
    -------
    i_t_r: float or list of floats
        total insulation of the clothing ensemble, [clo]

    """
    i_t = np.array(i_t)
    vr = np.array(vr)
    v_walk = np.array(v_walk)
    i_a_static = np.array(i_a_static)
    i_cl = np.array(i_cl)

    def normal_clothing(_vr, _vw, _i_t) -> float:
        return _i_t * _correction_normal_clothing(_vw=_vw, _vr=_vr)

    def nude(_vr, _vw, _i_a_static) -> float:
        return _i_a_static * _correction_nude(_vr=_vr, _vw=_vw)

    def low_clothing(_vr, _vw, _i_a_static, _i_cl, _i_t) -> float:
        return (
            (0.6 - _i_cl) * nude(_vr, _vw, _i_a_static)
            + _i_cl * normal_clothing(_vr, _vw, _i_t)
        ) / 0.6

    i_t_r = np.where(
        i_cl <= 0.6,
        low_clothing(_vr=vr, _vw=v_walk, _i_a_static=i_a_static, _i_cl=i_cl, _i_t=i_t),
        normal_clothing(_vr=vr, _vw=v_walk, _i_t=i_t),
    )
    return np.where(i_cl == 0, nude(_vr=vr, _vw=v_walk, _i_a_static=i_a_static), i_t_r)


def clo_correction_factor_environment(
    vr: float | list[float],
    v_walk: float | list[float],
    i_cl: float | list[float],
) -> float | list[float]:
    """Return the correction factor for the total insulation of the
    clothing ensemble (`I`:sub:`T`) or the basic/intrinsic insulation (`I`:sub:`cl`).

    This correction factor takes into account of the fact that the values of
    (`I`:sub:`T`) and (`I`:sub:`cl`) are estimated in static conditions. In real
    environments the person may be walking, activity may pump air through the clothing,
    etc.

    Parameters
    ----------
    vr: float or list of floats
        relative air speed, [m/s]
    v_walk: float or list of floats
        walking speed, [m/s]
    i_cl: float or list of floats
        intrinsic insulation of the clothing ensemble, this is the thermal insulation
        from the skin surface to the outer clothing surface [clo]

    Returns
    -------
    correction_factor: float or list of floats
        correction factor for the total insulation of the clothing ensemble
        (`I`:sub:`T,r` / (`I`:sub:`T`)) or the basic/intrinsic insulation
        (`I`:sub:`cl,r` / (`I`:sub:`cl`))

    """
    vr = np.array(vr)
    v_walk = np.array(v_walk)
    i_cl = np.array(i_cl)

    def correction_low_clothing(_vr, _vw, _i_cl) -> float:
        return (
            (0.6 - _i_cl) * _correction_nude(_vr, _vw)
            + _i_cl * _correction_normal_clothing(_vr, _vw)
        ) / 0.6

    c_f = np.where(
        i_cl <= 0.6,
        correction_low_clothing(_vr=vr, _vw=v_walk, _i_cl=i_cl),
        _correction_normal_clothing(_vr=vr, _vw=v_walk),
    )
    return np.where(i_cl == 0, _correction_nude(_vr=vr, _vw=v_walk), c_f)


def _correction_nude(_vr, _vw) -> float:
    """Calculate the correction factor for the total insulation of the clothing ensemble."""
    return np.exp(
        -0.533 * (_vr - 0.15)
        + 0.069 * (_vr - 0.15) ** 2
        - 0.462 * _vw
        + 0.201 * _vw**2,
    )


def _correction_normal_clothing(_vr, _vw) -> float:
    """Calculate the correction factor for normal clothing."""
    return np.exp(
        -0.281 * (_vr - 0.15)
        + 0.044 * (_vr - 0.15) ** 2
        - 0.492 * _vw
        + 0.176 * _vw**2,
    )


#: Met values of typical tasks.
met_typical_tasks = {
    "Sleeping": 0.7,
    "Reclining": 0.8,
    "Seated, quiet": 1.0,
    "Reading, seated": 1.0,
    "Writing": 1.0,
    "Typing": 1.1,
    "Standing, relaxed": 1.2,
    "Filing, seated": 1.2,
    "Flying aircraft, routine": 1.2,
    "Filing, standing": 1.4,
    "Driving a car": 1.5,
    "Walking about": 1.7,
    "Cooking": 1.8,
    "Table sawing": 1.8,
    "Walking 2mph (3.2kmh)": 2.0,
    "Lifting/packing": 2.1,
    "Seated, heavy limb movement": 2.2,
    "Light machine work": 2.2,
    "Flying aircraft, combat": 2.4,
    "Walking 3mph (4.8kmh)": 2.6,
    "House cleaning": 2.7,
    "Driving, heavy vehicle": 3.2,
    "Dancing": 3.4,
    "Calisthenics": 3.5,
    "Walking 4mph (6.4kmh)": 3.8,
    "Tennis": 3.8,
    "Heavy machine work": 4.0,
    "Handling 100lb (45 kg) bags": 4.0,
    "Pick and shovel work": 4.4,
    "Basketball": 6.3,
    "Wrestling": 7.8,
}

#: Total clothing insulation of typical ensembles.
clo_typical_ensembles = {
    "Walking shorts, short-sleeve shirt": 0.36,
    "Typical summer indoor clothing": 0.5,
    "Knee-length skirt, short-sleeve shirt, sandals, underwear": 0.54,
    "Trousers, short-sleeve shirt, socks, shoes, underwear": 0.57,
    "Trousers, long-sleeve shirt": 0.61,
    "Knee-length skirt, long-sleeve shirt, full slip": 0.67,
    "Sweat pants, long-sleeve sweatshirt": 0.74,
    "Jacket, Trousers, long-sleeve shirt": 0.96,
    "Typical winter indoor clothing": 1.0,
}

#: Clo values of individual clothing elements. To calculate the total
#: clothing insulation you need to add these values together.
clo_individual_garments = {
    "Metal chair": 0.00,
    "Bra": 0.01,
    "Wooden stool": 0.01,
    "Ankle socks": 0.02,
    "Shoes or sandals": 0.02,
    "Slippers": 0.03,
    "Panty hose": 0.02,
    "Calf length socks": 0.03,
    "Women's underwear": 0.03,
    "Men's underwear": 0.04,
    "Knee socks (thick)": 0.06,
    "Short shorts": 0.06,
    "Walking shorts": 0.08,
    "T-shirt": 0.08,
    "Standard office chair": 0.10,
    "Executive chair": 0.15,
    "Boots": 0.1,
    "Sleeveless scoop-neck blouse": 0.12,
    "Half slip": 0.14,
    "Long underwear bottoms": 0.15,
    "Full slip": 0.16,
    "Short-sleeve knit shirt": 0.17,
    "Sleeveless vest (thin)": 0.1,
    "Sleeveless vest (thick)": 0.17,
    "Sleeveless short gown (thin)": 0.18,
    "Short-sleeve dress shirt": 0.19,
    "Sleeveless long gown (thin)": 0.2,
    "Long underwear top": 0.2,
    "Thick skirt": 0.23,
    "Long-sleeve dress shirt": 0.25,
    "Long-sleeve flannel shirt": 0.34,
    "Long-sleeve sweat shirt": 0.34,
    "Short-sleeve hospital gown": 0.31,
    "Short-sleeve short robe (thin)": 0.34,
    "Short-sleeve pajamas": 0.42,
    "Long-sleeve long gown": 0.46,
    "Long-sleeve short wrap robe (thick)": 0.48,
    "Long-sleeve pajamas (thick)": 0.57,
    "Long-sleeve long wrap robe (thick)": 0.69,
    "Thin trousers": 0.15,
    "Thick trousers": 0.24,
    "Sweatpants": 0.28,
    "Overalls": 0.30,
    "Coveralls": 0.49,
    "Thin skirt": 0.14,
    "Long-sleeve shirt dress (thin)": 0.33,
    "Long-sleeve shirt dress (thick)": 0.47,
    "Short-sleeve shirt dress": 0.29,
    "Sleeveless, scoop-neck shirt (thin)": 0.23,
    "Sleeveless, scoop-neck shirt (thick)": 0.27,
    "Long sleeve shirt (thin)": 0.25,
    "Long sleeve shirt (thick)": 0.36,
    "Single-breasted coat (thin)": 0.36,
    "Single-breasted coat (thick)": 0.44,
    "Double-breasted coat (thin)": 0.42,
    "Double-breasted coat (thick)": 0.48,
}

#: This dictionary contains the reflection coefficients, Fr, for different
#: special materials
f_r_garments = {
    "Cotton with aluminium paint": 0.42,
    "Viscose with glossy aluminium foil": 0.19,
    "Aramid (Kevlar) with glossy aluminium foil": 0.14,
    "Wool with glossy aluminium foil": 0.12,
    "Cotton with glossy aluminium foil": 0.04,
    "Viscose vacuum metallized with aluminium": 0.06,
    "Aramid vacuum metallized with aluminium": 0.04,
    "Wool vacuum metallized with aluminium": 0.05,
    "Cotton vacuum metallized with aluminium": 0.05,
    "Glass fiber vacuum metallized with aluminium": 0.07,
}


class DefaultSkinTemperature(NamedTuple):
    """Default skin temperature in degree Celsius for 17 local body parts.

    The data comes from Hui Zhang's experiments
    https://escholarship.org/uc/item/3f4599hx
    """

    head: float = 35.3
    neck: float = 35.6
    chest: float = 35.1
    back: float = 35.3
    pelvis: float = 35.3
    left_shoulder: float = 34.2
    left_arm: float = 34.6
    left_hand: float = 34.4
    right_shoulder: float = 34.2
    right_arm: float = 34.6
    right_hand: float = 34.4
    left_thigh: float = 34.3
    left_leg: float = 32.8
    left_foot: float = 33.3
    right_thigh: float = 34.3
    right_leg: float = 32.8
    right_foot: float = 33.3



import numpy as np
import pytest

from pythermalcomfort.classes_return import PMVPPD
from pythermalcomfort.models import pmv_ppd_ashrae
from pythermalcomfort.utilities import Models
from tests.conftest import Urls, retrieve_reference_table, validate_result


def test_pmv_ppd(get_test_url, retrieve_data) -> None:
    """Test that the function calculates the PMV and PPD values correctly for various inputs."""
    reference_table = retrieve_reference_table(
        get_test_url,
        retrieve_data,
        Urls.PMV_PPD.name,
    )
    tolerance = reference_table["tolerance"]

    for entry in reference_table["data"]:
        inputs = entry["inputs"]
        outputs = entry["outputs"]
        # TODO change the validation table code and removed the following
        if "standard" not in inputs:
            inputs["standard"] = "iso"
        if inputs["standard"] == "ashrae":
            inputs["model"] = Models.ashrae_55_2023.value
            del inputs["standard"]
            result = pmv_ppd_ashrae(**inputs)

            validate_result(result, outputs, tolerance)


class TestPmvPpd:
    """Test cases for the PMV and PPD model."""

    def test_thermal_sensation(self) -> None:
        """Test that the function returns the correct thermal sensation values."""
        np.testing.assert_equal(
            pmv_ppd_ashrae(
                [16, 21, 24, 26, 29, 32, 34, 33.47, 33.46],
                [16, 21, 24, 26, 29, 32, 34, 33.47, 33.46],
                0.2,
                50,
                1,
                0.5,
                model=Models.ashrae_55_2023.value,
            ).tsv,
            [
                "Cold",
                "Cool",
                "Slightly Cool",
                "Neutral",
                "Slightly Warm",
                "Warm",
                "Hot",
                "Hot",
                "Warm",
            ],
        )

    #  Returns NaN for invalid input values
    def test_returns_nan_for_invalid_input_values(self) -> None:
        """Test that the function returns NaN for invalid input values."""
        # test airspeed limits
        np.testing.assert_equal(
            pmv_ppd_ashrae(
                [26, 24, 22, 26, 24, 22],
                [26, 24, 22, 26, 24, 22],
                [0.9, 0.6, 0.3, 0.9, 0.6, 0.3],
                50,
                [1.1, 1.1, 1.1, 1.3, 1.3, 1.3],
                [0.5, 0.5, 0.5, 0.7, 0.7, 0.7],
                model=Models.ashrae_55_2023.value,
                airspeed_control=False,
            ).pmv,
            [np.nan, np.nan, np.nan, -0.14, -0.43, -0.57],
        )

        np.testing.assert_equal(
            pmv_ppd_ashrae(
                [41, 20, 20, 20, 20, 39],
                [20, 41, 20, 20, 20, 39],
                [0.1, 0.1, 2.1, 0.1, 0.1, 0.1],
                50,
                [1.1, 1.1, 1.1, 0.7, 1.1, 3.9],
                [0.5, 0.5, 0.5, 0.5, 2.1, 1.9],
                model=Models.ashrae_55_2023.value,
            ).pmv,
            np.array([np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]),
        )

        np.testing.assert_equal(
            pmv_ppd_ashrae(
                41,
                41,
                2,
                50,
                0.7,
                2.1,
                model=Models.ashrae_55_2023.value,
                limit_inputs=False,
            ),
            PMVPPD(pmv=np.float64(4.48), ppd=np.float64(100.0), tsv=np.str_("Hot")),
        )

    def test_wrong_standard(self) -> None:
        """Test that the function raises a ValueError for an unsupported model."""
        with pytest.raises(ValueError):
            pmv_ppd_ashrae(25, 25, 0.1, 50, 1.1, 0.5, model="random")



from pprint import pprint

from pythermalcomfort.models import adaptive_ashrae
from pythermalcomfort.utilities import running_mean_outdoor_temperature

result = adaptive_ashrae(tdb=25, tr=25, t_running_mean=23, v=0.3)

pprint(result)

print(result.acceptability_80)  # or use result["acceptability_80"]

result = adaptive_ashrae(tdb=77, tr=77, t_running_mean=73.5, v=1, units="IP")

pprint(result)

print(result["acceptability_80"])

rmt_value = running_mean_outdoor_temperature([29, 28, 30, 29, 28, 30, 27], alpha=0.9)

result = adaptive_ashrae(tdb=25, tr=25, t_running_mean=rmt_value, v=0.3)
pprint(result)
